# Copyright 2012-2024 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Ssim class of bluecellulab that loads a circuit simulation to do cell
simulations."""


from __future__ import annotations
from collections.abc import Iterable
from pathlib import Path
from typing import Optional
import logging

import neuron
import numpy as np
import pandas as pd
from pydantic.types import NonNegativeInt
import bluecellulab
from bluecellulab.cell import CellDict
from bluecellulab.cell.sonata_proxy import SonataProxy
from bluecellulab.circuit import CellId, SimulationValidator, SynapseProperty
from bluecellulab.circuit.circuit_access import (
    CircuitAccess,
    BluepyCircuitAccess,
    SonataCircuitAccess,
    get_synapse_connection_parameters
)
from bluecellulab.circuit.config import SimulationConfig
from bluecellulab.circuit.format import determine_circuit_format, CircuitFormat
from bluecellulab.circuit.node_id import create_cell_id, create_cell_ids
from bluecellulab.circuit.simulation_access import BluepySimulationAccess, SimulationAccess, SonataSimulationAccess, _sample_array
from bluecellulab.importer import load_hoc_and_mod_files
from bluecellulab.rngsettings import RNGSettings
from bluecellulab.stimulus.circuit_stimulus_definitions import Noise, OrnsteinUhlenbeck, RelativeOrnsteinUhlenbeck, RelativeShotNoise, ShotNoise
import bluecellulab.stimulus.circuit_stimulus_definitions as circuit_stimulus_definitions
from bluecellulab.exceptions import BluecellulabError
from bluecellulab.simulation import (
    set_global_condition_parameters,
    set_tstop_value
)
from bluecellulab.synapse.synapse_types import SynapseID

logger = logging.getLogger(__name__)


class SSim:
    """Class that loads a circuit simulation to do cell simulations."""

    @load_hoc_and_mod_files
    def __init__(
        self,
        simulation_config: str | Path | SimulationConfig,
        dt: float = 0.025,
        record_dt: Optional[float] = None,
        base_seed: Optional[NonNegativeInt] = None,
        rng_mode: Optional[str] = None,
        print_cellstate: bool = False,
    ):
        """

        Parameters
        ----------
        simulation_config : Absolute filename of the simulation config file.
        dt : Timestep of the simulation
        record_dt : Sampling interval of the recordings
        base_seed : Base seed used for this simulation. Setting this
                    will override the value set in the simulation config.
        rng_mode : String with rng mode, if not specified mode is taken from
                    simulation config. Possible values are Compatibility, Random123
                    and UpdatedMCell.
        print_cellstate:
                    Flag to use NEURON prcellstate for simulation GIDs
        """
        self.dt = dt
        self.record_dt = record_dt

        self.circuit_format = determine_circuit_format(simulation_config)
        if self.circuit_format == CircuitFormat.SONATA:
            self.circuit_access: CircuitAccess = SonataCircuitAccess(simulation_config)
            self.simulation_access: SimulationAccess = SonataSimulationAccess(simulation_config)
        else:
            self.circuit_access = BluepyCircuitAccess(simulation_config)
            self.simulation_access = BluepySimulationAccess(simulation_config)
            SimulationValidator(self.circuit_access).validate()

        self.pc = neuron.h.ParallelContext() if print_cellstate else None

        self.rng_settings = RNGSettings.get_instance()
        self.rng_settings.set_seeds(
            rng_mode,
            self.circuit_access.config,
            base_seed=base_seed
        )

        self.cells: CellDict = CellDict()

        self.gids_instantiated = False

        # Make sure tstop is set correctly, because it is used by the
        # TStim noise stimulus
        if self.circuit_access.config.duration is not None:
            set_tstop_value(self.circuit_access.config.duration)

        self.spike_threshold = self.circuit_access.config.spike_threshold
        self.spike_location = self.circuit_access.config.spike_location

        condition_parameters = self.circuit_access.config.condition_parameters()
        set_global_condition_parameters(condition_parameters)

    def instantiate_gids(
        self,
        cells: int | tuple[str, int] | list[int] | list[tuple[str, int]],
        add_replay: bool = False,
        add_stimuli: bool = False,
        add_synapses: bool = False,
        add_minis: bool = False,
        add_noise_stimuli: bool = False,
        add_hyperpolarizing_stimuli: bool = False,
        add_relativelinear_stimuli: bool = False,
        add_pulse_stimuli: bool = False,
        add_projections: bool | list[str] = False,
        intersect_pre_gids: Optional[list] = None,
        interconnect_cells: bool = True,
        pre_spike_trains: None | dict[tuple[str, int], Iterable] | dict[int, Iterable] = None,
        add_shotnoise_stimuli: bool = False,
        add_ornstein_uhlenbeck_stimuli: bool = False,
    ):
        """Instantiate a list of cells.

        Parameters
        ----------
        cells :
               List of cell ids. When a single element, it will be converted to a list
        add_replay : Add presynaptic spiketrains from the large simulation
                     If pre_spike_trains is combined with this option the
                     spiketrains will be merged
        add_stimuli : Add the same stimuli as in the large simulation
        add_synapses :
                       Add the touch-detected synapses, as described by the
                       circuit to the cell
                       (This option only influence the 'creation' of synapses,
                       it doesn't add any connections)
                       Default value is False
        add_minis : Add synaptic minis to the synapses
                    (this requires add_synapses=True)
                    Default value is False
        add_noise_stimuli :
                            Process the 'noise' stimuli blocks of the
                            simulation config,
                            Setting add_stimuli=True,
                            will automatically set this option to True.
        add_hyperpolarizing_stimuli : Process the 'hyperpolarizing' stimuli
                                      blocks of the simulation config.
                                      Setting add_stimuli=True,
                                      will automatically set this option to
                                      True.
        add_relativelinear_stimuli : Process the 'relativelinear' stimuli
                                     blocks of the simulation config.
                                     Setting add_stimuli=True,
                                     will automatically set this option to
                                     True.
        add_pulse_stimuli : Process the 'pulse' stimuli
                            blocks of the simulation config.
                            Setting add_stimuli=True,
                            will automatically set this option to
                            True.
        add_projections:
                         If True, adds all of the projection blocks of the
                         circuit config. If False, no projections are added.
                         If list, adds only the projections in the list.
        intersect_pre_gids : list of gids
                             Only add synapses to the cells if their
                             presynaptic gid is in this list
        interconnect_cells : When multiple gids are instantiated,
                             interconnect the cells with real (non-replay)
                             synapses. When this option is combined with
                             add_replay, replay spiketrains will only be added
                             for those presynaptic cells that are not in the
                             network that's instantiated.
                             This option requires add_synapses=True
        pre_spike_trains : A dictionary with keys the presynaptic gids, and
                           values the list of spike timings of the
                           presynaptic cells with the given gids.
                           If this option is used in combination with
                           add_replay=True, the spike trains for the same
                           gids will be automatically merged
        add_shotnoise_stimuli :
                            Process the 'shotnoise' stimuli blocks of the
                            simulation config,
                            Setting add_stimuli=True,
                            will automatically set this option to True.
        add_ornstein_uhlenbeck_stimuli :
                            Process the 'ornstein_uhlenbeck' stimuli blocks
                            of the simulation config,
                            Setting add_stimuli=True,
                            will automatically set this option to True.
        """
        if not isinstance(cells, Iterable) or isinstance(cells, tuple):
            cells = [cells]

        # convert to CellId objects
        cell_ids: list[CellId] = create_cell_ids(cells)
        if intersect_pre_gids is not None:
            pre_gids: Optional[list[CellId]] = create_cell_ids(intersect_pre_gids)
        else:
            pre_gids = None

        # if pre_spike_trains take int as key then convert to CellId
        if pre_spike_trains is not None:
            if not isinstance(next(iter(pre_spike_trains.keys())), tuple):
                pre_spike_trains = {
                    create_cell_id(gid): pre_spike_trains[gid]  # type: ignore
                    for gid in pre_spike_trains
                }

        if self.gids_instantiated:
            raise Exception("SSim: instantiate_gids() called twice on the \
                    same SSim, this is not supported yet")
        else:
            self.gids_instantiated = True

        if pre_spike_trains or add_replay:
            if add_synapses is False:
                raise Exception("SSim: you need to set add_synapses to True "
                                "if you want to specify use add_replay or "
                                "pre_spike_trains")

        if add_projections is True:
            projections = self.circuit_access.config.get_all_projection_names()
        elif add_projections is False:
            projections = []
        else:
            projections = add_projections

        self._add_cells(cell_ids)
        if add_synapses:
            self._add_synapses(
                pre_gids=pre_gids,
                add_minis=add_minis,
                projections=projections)
        if add_replay or interconnect_cells or pre_spike_trains:
            if add_replay and not add_synapses:
                raise Exception("SSim: add_replay option can not be used if "
                                "add_synapses is False")
            self._add_connections(add_replay=add_replay,
                                  interconnect_cells=interconnect_cells,
                                  user_pre_spike_trains=pre_spike_trains)  # type: ignore
        if add_stimuli:
            add_noise_stimuli = True
            add_hyperpolarizing_stimuli = True
            add_relativelinear_stimuli = True
            add_pulse_stimuli = True
            add_shotnoise_stimuli = True
            add_ornstein_uhlenbeck_stimuli = True

        if add_noise_stimuli or \
                add_hyperpolarizing_stimuli or \
                add_pulse_stimuli or \
                add_relativelinear_stimuli or \
                add_shotnoise_stimuli or \
                add_ornstein_uhlenbeck_stimuli:
            self._add_stimuli(
                add_noise_stimuli=add_noise_stimuli,
                add_hyperpolarizing_stimuli=add_hyperpolarizing_stimuli,
                add_relativelinear_stimuli=add_relativelinear_stimuli,
                add_pulse_stimuli=add_pulse_stimuli,
                add_shotnoise_stimuli=add_shotnoise_stimuli,
                add_ornstein_uhlenbeck_stimuli=add_ornstein_uhlenbeck_stimuli
            )

    def _add_stimuli(self, add_noise_stimuli=False,
                     add_hyperpolarizing_stimuli=False,
                     add_relativelinear_stimuli=False,
                     add_pulse_stimuli=False,
                     add_shotnoise_stimuli=False,
                     add_ornstein_uhlenbeck_stimuli=False
                     ):
        """Instantiate all the stimuli."""
        stimuli_entries = self.circuit_access.config.get_all_stimuli_entries()
        # Also add the injections / stimulations as in the cortical model
        # check in which StimulusInjects the gid is a target
        # Every noise or shot noise stimulus gets a new seed
        noisestim_count = 0
        shotnoise_stim_count = 0
        ornstein_uhlenbeck_stim_count = 0

        for stimulus in stimuli_entries:
            target = stimulus.target
            gids_of_target = self.circuit_access.get_target_cell_ids(target)

            for cell_id in self.cells:
                if cell_id not in gids_of_target:
                    continue
                if isinstance(stimulus, circuit_stimulus_definitions.Noise):
                    if add_noise_stimuli:
                        self.cells[cell_id].add_replay_noise(
                            stimulus, noisestim_count=noisestim_count)
                elif isinstance(stimulus, circuit_stimulus_definitions.Hyperpolarizing):
                    if add_hyperpolarizing_stimuli:
                        self.cells[cell_id].add_replay_hypamp(stimulus)
                elif isinstance(stimulus, circuit_stimulus_definitions.Pulse):
                    if add_pulse_stimuli:
                        self.cells[cell_id].add_pulse(stimulus)
                elif isinstance(stimulus, circuit_stimulus_definitions.RelativeLinear):
                    if add_relativelinear_stimuli:
                        self.cells[cell_id].add_replay_relativelinear(stimulus)
                elif isinstance(stimulus, circuit_stimulus_definitions.ShotNoise):
                    if add_shotnoise_stimuli:
                        self.cells[cell_id].add_replay_shotnoise(
                            self.cells[cell_id].soma, 0.5, stimulus,
                            shotnoise_stim_count=shotnoise_stim_count)
                elif isinstance(stimulus, circuit_stimulus_definitions.RelativeShotNoise):
                    if add_shotnoise_stimuli:
                        self.cells[cell_id].add_replay_relative_shotnoise(
                            self.cells[cell_id].soma, 0.5, stimulus,
                            shotnoise_stim_count=shotnoise_stim_count)
                elif isinstance(stimulus, circuit_stimulus_definitions.OrnsteinUhlenbeck):
                    if add_ornstein_uhlenbeck_stimuli:
                        self.cells[cell_id].add_ornstein_uhlenbeck(
                            self.cells[cell_id].soma, 0.5, stimulus,
                            stim_count=ornstein_uhlenbeck_stim_count)
                elif isinstance(stimulus, circuit_stimulus_definitions.RelativeOrnsteinUhlenbeck):
                    if add_ornstein_uhlenbeck_stimuli:
                        self.cells[cell_id].add_relative_ornstein_uhlenbeck(
                            self.cells[cell_id].soma, 0.5, stimulus,
                            stim_count=ornstein_uhlenbeck_stim_count)
                elif isinstance(stimulus, circuit_stimulus_definitions.SynapseReplay):  # sonata only
                    if self.circuit_access.target_contains_cell(
                        stimulus.target, cell_id
                    ):
                        self.cells[cell_id].add_synapse_replay(
                            stimulus, self.spike_threshold, self.spike_location
                        )
                else:
                    raise ValueError("Found stimulus with pattern %s, "
                                     "not supported" % stimulus)
                logger.debug(f"Added {stimulus} to cell_id {cell_id}")

            if isinstance(stimulus, Noise):
                noisestim_count += 1
            elif isinstance(stimulus, (ShotNoise, RelativeShotNoise)):
                shotnoise_stim_count += 1
            elif isinstance(stimulus, (OrnsteinUhlenbeck, RelativeOrnsteinUhlenbeck)):
                ornstein_uhlenbeck_stim_count += 1

    def _add_synapses(
            self, pre_gids=None, add_minis=False, projections=None):
        """Instantiate all the synapses."""
        for cell_id in self.cells:
            self._add_cell_synapses(
                cell_id, pre_gids=pre_gids,
                add_minis=add_minis,
                projections=projections)

    def _add_cell_synapses(
        self, cell_id: CellId, pre_gids=None, add_minis=False, projections=None
    ) -> None:
        syn_descriptions = self.get_syn_descriptions(cell_id, projections=projections)

        if pre_gids is not None:
            if self.circuit_format == CircuitFormat.SONATA:
                syn_descriptions = self._intersect_pre_gids_cell_ids_multipopulation(
                    syn_descriptions, pre_gids
                )
            else:
                syn_descriptions = self._intersect_pre_gids(syn_descriptions, pre_gids)

        # Check if there are any presynaptic cells, otherwise skip adding
        # synapses
        if syn_descriptions.empty:
            logger.warning(
                f"No presynaptic cells found for gid {cell_id}, no synapses added"
            )

        else:
            for idx, syn_description in syn_descriptions.iterrows():
                popids = (
                    syn_description["source_popid"],
                    syn_description["target_popid"],
                )
                self._instantiate_synapse(
                    cell_id=cell_id,
                    syn_id=idx,  # type: ignore
                    syn_description=syn_description,
                    add_minis=add_minis,
                    popids=popids
                )
            logger.info(f"Added {syn_descriptions} synapses for gid {cell_id}")
            if add_minis:
                logger.info(f"Added minis for {cell_id=}")

    @staticmethod
    def _intersect_pre_gids(syn_descriptions, pre_gids: list[CellId]) -> pd.DataFrame:
        """Return the synapse descriptions with pre_gids intersected."""
        _pre_gids = {x.id for x in pre_gids}
        return syn_descriptions[syn_descriptions[SynapseProperty.PRE_GID].isin(_pre_gids)]

    @staticmethod
    def _intersect_pre_gids_cell_ids_multipopulation(syn_descriptions, pre_cell_ids: list[CellId]) -> pd.DataFrame:
        """Return the synapse descriptions with pre_cell_ids intersected.

        Supports multipopulations.
        """
        filtered_rows = syn_descriptions.apply(
            lambda row: any(
                cell.population_name == row["source_population_name"] and row[SynapseProperty.PRE_GID] == cell.id
                for cell in pre_cell_ids
            ),
            axis=1,
        )
        return syn_descriptions[filtered_rows]

    def get_syn_descriptions(
        self, cell_id: int | tuple[str, int], projections=None
    ) -> pd.DataFrame:
        """Get synapse descriptions dataframe."""
        cell_id = create_cell_id(cell_id)
        return self.circuit_access.extract_synapses(cell_id, projections=projections)

    @staticmethod
    def merge_pre_spike_trains(*train_dicts) -> dict[CellId, np.ndarray]:
        """Merge presynaptic spike train dicts."""
        filtered_dicts = [d for d in train_dicts if d not in [None, {}, [], ()]]

        all_keys = set().union(*[d.keys() for d in filtered_dicts])
        return {
            k: np.sort(np.concatenate([d[k] for d in filtered_dicts if k in d]))
            for k in all_keys
        }

    def _add_connections(
            self,
            add_replay=None,
            interconnect_cells=None,
            user_pre_spike_trains: None | dict[CellId, Iterable] = None) -> None:
        """Instantiate the (replay and real) connections in the network."""
        pre_spike_trains = self.simulation_access.get_spikes() if add_replay else {}
        pre_spike_trains = self.merge_pre_spike_trains(
            pre_spike_trains,
            user_pre_spike_trains)

        for post_gid in self.cells:
            for syn_id in self.cells[post_gid].synapses:
                synapse = self.cells[post_gid].synapses[syn_id]
                syn_description: pd.Series = synapse.syn_description
                delay_weights = synapse.delay_weights
                source_population = syn_description["source_population_name"]
                pre_gid = CellId(source_population, int(syn_description[SynapseProperty.PRE_GID]))

                real_synapse_connection = pre_gid in self.cells and interconnect_cells

                if real_synapse_connection:
                    if (
                            user_pre_spike_trains is not None
                            and pre_gid in user_pre_spike_trains
                    ):
                        raise BluecellulabError(
                            """Specifying prespike trains of real connections"""
                            """ is not allowed."""
                        )
                    connection = bluecellulab.Connection(
                        self.cells[post_gid].synapses[syn_id],
                        pre_spiketrain=None,
                        pre_cell=self.cells[pre_gid],
                        stim_dt=self.dt,
                        parallel_context=self.pc,
                        spike_threshold=self.spike_threshold,
                        spike_location=self.spike_location)

                    logger.debug(f"Added real connection between {pre_gid} and {post_gid}, {syn_id}")
                else:  # replay connection
                    try:
                        pre_spiketrain = pre_spike_trains[pre_gid]
                    except KeyError:
                        pre_spiketrain = None

                    connection = bluecellulab.Connection(
                        self.cells[post_gid].synapses[syn_id],
                        pre_spiketrain=pre_spiketrain,
                        pre_cell=None,
                        stim_dt=self.dt,
                        spike_threshold=self.spike_threshold,
                        spike_location=self.spike_location)
                    logger.debug(f"Added replay connection from {pre_gid} to {post_gid}, {syn_id}")

                self.cells[post_gid].connections[syn_id] = connection
                for delay, weight_scale in delay_weights:
                    self.cells[post_gid].add_replay_delayed_weight(
                        syn_id, delay,
                        weight_scale * connection.weight)

            if len(self.cells[post_gid].connections) > 0:
                logger.debug(f"Added synaptic connections for target {post_gid}")

    def _add_cells(self, cell_ids: list[CellId]) -> None:
        """Instantiate cells from a gid list."""
        self.cells = CellDict()

        for cell_id in cell_ids:
            self.cells[cell_id] = cell = self.create_cell_from_circuit(cell_id)
            if self.circuit_access.node_properties_available:
                cell.connect_to_circuit(SonataProxy(cell_id, self.circuit_access))
            if self.pc is not None:
                self.pc.set_gid2node(cell_id.id, self.pc.id())  # register GID for this node
                nc = self.cells[cell_id].create_netcon_spikedetector(
                    None, location=self.spike_location, threshold=self.spike_threshold)
                self.pc.cell(cell_id.id, nc)  # register cell spike detector

    def _instantiate_synapse(self, cell_id: CellId, syn_id: SynapseID, syn_description,
                             add_minis=False, popids=(0, 0)) -> None:
        """Instantiate one synapse for a given gid, syn_id and
        syn_description."""
        pre_cell_id = CellId(cell_id.population_name, int(syn_description[SynapseProperty.PRE_GID]))
        syn_connection_parameters = get_synapse_connection_parameters(
            circuit_access=self.circuit_access,
            pre_cell=pre_cell_id,
            post_cell=cell_id)
        if syn_connection_parameters["add_synapse"]:
            condition_parameters = self.circuit_access.config.condition_parameters()

            self.cells[cell_id].add_replay_synapse(
                syn_id, syn_description, syn_connection_parameters, condition_parameters,
                popids=popids, extracellular_calcium=self.circuit_access.config.extracellular_calcium)
            if add_minis:
                mini_frequencies = self.circuit_access.fetch_mini_frequencies(cell_id)
                logger.debug(f"Adding minis for synapse {syn_id}: syn_description={syn_description}, connection={syn_connection_parameters}, frequency={mini_frequencies}")

                self.cells[cell_id].add_replay_minis(
                    syn_id,
                    syn_description,
                    syn_connection_parameters,
                    popids=popids,
                    mini_frequencies=mini_frequencies,
                )

    def run(
        self,
        t_stop: Optional[float] = None,
        v_init: Optional[float] = None,
        celsius: Optional[float] = None,
        dt: Optional[float] = None,
        forward_skip: bool = True,
        forward_skip_value: Optional[float] = None,
        cvode: bool = False,
        show_progress: bool = False,
    ):
        """Simulate the SSim.

        Parameters
        ----------
        t_stop :
            This function will run the simulation until t_stop
        v_init :
            Voltage initial value when the simulation starts
        celsius :
            Temperature at which the simulation runs
        dt :
            Timestep (delta-t) for the simulation
        forward_skip :
                       [compatibility/non-sonata] Enable/disable ForwardSkip, when
                       forward_skip_value is None, forward skip will only be
                       enabled if the simulation config has a ForwardSkip value)
        forward_skip_value :
                       [compatibility/non-sonata] Overwrite the ForwardSkip value
                       in the simulation config. If this is set to None, the value
                       in the simulation config is used.
        cvode :
                Force the simulation to run in variable timestep. Not possible
                when there are stochastic channels in the neuron model. When
                enabled results from a large network simulation will not be
                exactly reproduced.
        show_progress:
                       Show a progress bar during simulations. When
                       enabled results from a large network simulation
                       will not be exactly reproduced.
        """
        if t_stop is None:
            duration = self.circuit_access.config.duration
            if duration is None:  # type narrowing
                t_stop = 0.0
            else:
                t_stop = duration
        if dt is None:
            dt = self.circuit_access.config.dt
        if forward_skip_value is None:
            forward_skip_value = self.circuit_access.config.forward_skip
        if celsius is None:
            celsius = self.circuit_access.config.celsius
        if v_init is None:
            v_init = self.circuit_access.config.v_init

        sim = bluecellulab.Simulation(self.pc)
        for cell_id in self.cells:
            sim.add_cell(self.cells[cell_id])

        if show_progress:
            logger.warning("show_progress enabled, this will very likely"
                           "break the exact reproducibility of large network"
                           "simulations")

        sim.run(
            t_stop,
            cvode=cvode,
            dt=dt,
            celsius=celsius,
            v_init=v_init,
            forward_skip=forward_skip,
            forward_skip_value=forward_skip_value,
            show_progress=show_progress)

    def get_mainsim_voltage_trace(
            self, cell_id: int | tuple[str, int], t_start=None, t_stop=None, t_step=None
    ) -> np.ndarray:
        """Get the voltage trace from a cell from the main simulation.

        Parameters
        -----------
        cell_id: cell id of interest.
        t_start, t_stop: time range of interest,
        report time range is used by default.
        t_step: time step (should be a multiple of report time step T;
        equals T by default)

        Returns:
            One dimentional np.ndarray to represent the voltages.
        """
        cell_id = create_cell_id(cell_id)
        return self.simulation_access.get_soma_voltage(cell_id, t_start, t_stop, t_step)

    def get_mainsim_time_trace(self, t_step=None) -> np.ndarray:
        """Get the time trace from the main simulation.

        Parameters
        -----------
        t_step: time step (should be a multiple of report time step T;
        equals T by default)

        Returns:
            One dimentional np.ndarray to represent the times.
        """
        return self.simulation_access.get_soma_time_trace(t_step)

    def get_time(self) -> np.ndarray:
        """Get the time vector for the recordings, contains negative times.

        The negative times occur as a result of ForwardSkip.
        """
        first_key = next(iter(self.cells))
        return self.cells[first_key].get_time()

    def get_time_trace(self, t_step=None) -> np.ndarray:
        """Get the time vector for the recordings, negative times removed.

        Parameters
        -----------
        t_step: time step (should be a multiple of report time step T;
        equals T by default)

        Returns:
            One dimentional np.ndarray to represent the times.
        """
        time = self.get_time()
        time = time[np.where(time >= 0.0)]

        if t_step is not None:
            ratio = t_step / self.dt
            time = _sample_array(time, ratio)
        return time

    def get_voltage_trace(
            self, cell_id: int | tuple[str, int], t_start=None, t_stop=None, t_step=None
    ) -> np.ndarray:
        """Get the voltage vector for the cell_id, negative times removed.

        Parameters
        -----------
        cell_id: cell id of interest.
        t_start, t_stop: time range of interest,
        report time range is used by default.
        t_step: time step (should be a multiple of report time step T;
        equals T by default)

        Returns:
            One dimentional np.ndarray to represent the voltages.
        """
        cell_id = create_cell_id(cell_id)
        time = self.get_time()
        voltage = self.cells[cell_id].get_soma_voltage()

        if t_start is None or t_start < 0:
            t_start = 0
        if t_stop is None:
            t_stop = np.inf

        voltage = voltage[np.where((time >= t_start) & (time <= t_stop))]

        if t_step is not None:
            ratio = t_step / self.dt
            voltage = _sample_array(voltage, ratio)
        return voltage

    def delete(self):
        """Delete ssim and all of its attributes.

        NEURON objects are explicitly needed to be deleted.
        """
        if hasattr(self, 'cells'):
            for _, cell in self.cells.items():
                cell.delete()
            cell_ids = list(self.cells.keys())
            for cell_id in cell_ids:
                del self.cells[cell_id]

    def __del__(self):
        """Destructor.

        Deletes all allocated NEURON objects.
        """
        self.delete()

    def fetch_cell_kwargs(self, cell_id: CellId) -> dict:
        """Get the kwargs to instantiate a Cell object."""
        emodel_properties = self.circuit_access.get_emodel_properties(cell_id)
        cell_kwargs = {
            'template_path': self.circuit_access.emodel_path(cell_id),
            'morphology_path': self.circuit_access.morph_filepath(cell_id),
            'cell_id': cell_id,
            'record_dt': self.record_dt,
            'template_format': self.circuit_access.get_template_format(),
            'emodel_properties': emodel_properties,
        }

        return cell_kwargs

    def create_cell_from_circuit(self, cell_id: CellId) -> bluecellulab.Cell:
        """Create a Cell object from the circuit."""
        cell_kwargs = self.fetch_cell_kwargs(cell_id)
        return bluecellulab.Cell(template_path=cell_kwargs['template_path'],
                                 morphology_path=cell_kwargs['morphology_path'],
                                 cell_id=cell_kwargs['cell_id'],
                                 record_dt=cell_kwargs['record_dt'],
                                 template_format=cell_kwargs['template_format'],
                                 emodel_properties=cell_kwargs['emodel_properties'])



"""Unit tests for cell/random module."""
import neuron
from pytest import approx

from bluecellulab.cell.random import gamma


def test_gamma():
    """Unit test for the gamma function."""
    rng = neuron.h.Random()
    gamma_shape = 0.5
    gamma_scale = 1.5
    N = 5
    res = gamma(rng, gamma_shape, gamma_scale, N)
    assert len(res) == N
    assert sum(res) == approx(2.9341513)
    assert res[0] == approx(0.2862183)
    assert max(res) == approx(1.3015527)
    assert min(res) == approx(0.2802995)



"""Package for performing single cell or network simulations and
experiments."""

try:
    import bluepy
    BLUEPY_AVAILABLE = True
except ImportError:
    BLUEPY_AVAILABLE = False

from .importer import *  # NOQA
from .tools import *  # NOQA
from .verbosity import *
from .cell import Cell, create_ball_stick  # NOQA
from .connection import Connection  # NOQA
from .plotwindow import PlotWindow  # NOQA
from .dendrogram import Dendrogram  # NOQA
from .psection import PSection  # NOQA
from .psegment import PSegment  # NOQA
from .simulation import Simulation  # NOQA
from .rngsettings import RNGSettings  # NOQA
from .ssim import SSim  # NOQA



"""Tests for the Simulation class"""

import pathlib

import numpy as np
from pytest import approx
import pytest

import bluecellulab

parent_dir = pathlib.Path(__file__).resolve().parent.parent


@pytest.mark.v5
class TestCellBaseClass:
    """Base test class"""

    def setup_method(self):
        """Setup"""
        self.cell = bluecellulab.Cell(
            parent_dir / "examples/cell_example1/test_cell.hoc",
            str(parent_dir / "examples/cell_example1"))
        self.sim = bluecellulab.Simulation()
        self.sim.add_cell(self.cell)
        assert isinstance(self.sim, bluecellulab.Simulation)

    def teardown_method(self):
        """Destructor"""
        del self.cell
        del self.sim

    def test_run(self):
        """Simulation: Run the simulation for 20 ms"""
        self.sim.run(20)


@pytest.mark.v5
class TestCellcSTUTRandom123BaseClass:
    """Base test class"""

    def setup_method(self):
        """Setup"""
        self.cell = bluecellulab.Cell(
            parent_dir / "examples/cell_example_cstut/cSTUT_7.hoc",
            str(parent_dir / "examples/cell_example_cstut"))
        self.sim = bluecellulab.Simulation()
        self.sim.add_cell(self.cell)
        assert isinstance(self.sim, bluecellulab.Simulation)

    def teardown_method(self):
        """Destructor"""
        del self.cell
        del self.sim

    def test_run(self):
        """Simulation: Run a simulation with cSTUT and Random123 for 300 ms"""
        self.sim.run(300, cvode=False, use_random123_stochkv=True)

        time = self.cell.get_time()
        voltage = self.cell.get_soma_voltage()

        voltage_ss = voltage[np.where(time > 150)]

        # Lowered precision because of
        # commit 81a7a398214f2f5fba199ac3672c3dc3ccb6b103
        # in nrn repo
        # self.cell = bluecellulab.Cell("%s/examples/cell_example_cstut/cSTUT_7.hoc" % script_dir,"%s/examples/cell_example_cstut" % script_dir)
        assert np.mean(voltage_ss) == approx(-75.5400762008, abs=1e-6)
        assert np.std(voltage_ss) == approx(0.142647101877, abs=1e-6)


@pytest.mark.v5
class TestCellcSTUTBaseClass:

    """Base test class"""

    def setup_method(self):
        """Setup"""
        self.cell = bluecellulab.Cell(
            parent_dir / "examples/cell_example_cstut/cSTUT_7.hoc",
            str(parent_dir / "examples/cell_example_cstut"))
        self.sim = bluecellulab.Simulation()
        self.sim.add_cell(self.cell)
        assert isinstance(self.sim, bluecellulab.Simulation)

    def teardown_method(self):
        """Destructor"""
        del self.cell
        del self.sim

    def test_run(self):
        """Simulation: Run a simulation with cSTUT and MCellRan4 for 300 ms"""
        self.sim.run(300, cvode=False, use_random123_stochkv=False)

        time = self.cell.get_time()
        voltage = self.cell.get_soma_voltage()

        voltage_ss = voltage[np.where(time > 150)]

        # Lowered precision because of
        # commit 81a7a398214f2f5fba199ac3672c3dc3ccb6b103
        # in nrn repo
        assert np.mean(voltage_ss) == approx(-75.61918061202924, abs=1e-6)
        assert np.std(voltage_ss) == approx(0.19192736450671288, abs=1e-6)



"""Unit tests for circuit/config/simulation_config.py."""

from pathlib import Path

from bluepysnap import Simulation as SnapSimulation
import pytest

from bluecellulab.circuit.config import SonataSimulationConfig
from bluecellulab.circuit.config.sections import (
    ConditionEntry,
    Conditions,
    ConnectionOverrides,
    MechanismConditions,
)
from bluecellulab.stimulus.circuit_stimulus_definitions import Noise, Hyperpolarizing
from tests.helpers.os_utils import cwd


parent_dir = Path(__file__).resolve().parent.parent
cond_params_conf_path = (
    parent_dir
    / "examples"
    / "sonata_unit_test_sims"
    / "condition_parameters"
    / "simulation_config.json"
)

multi_input_conf_path = (
    parent_dir
    / "examples"
    / "sonata_unit_test_sims"
    / "condition_parameters"
    / "simulation_config_many_inputs.json"
)

hipp_sim_with_projection = (
    parent_dir
    / "examples"
    / "sonata_unit_test_sims"
    / "projections"
    / "simulation_config.json"
)


def test_non_existing_config():
    with pytest.raises(FileNotFoundError):
        SonataSimulationConfig(parent_dir / "examples" / "non_existing_config")


def test_init_with_snap_simulation_config():
    snap_sim = SnapSimulation(cond_params_conf_path)
    SonataSimulationConfig(snap_sim)


def test_init_with_invalid_type():
    with pytest.raises(TypeError):
        SonataSimulationConfig({"invalid": "type", "dict": "config"})


def test_get_all_stimuli_entries():
    sim = SonataSimulationConfig(multi_input_conf_path)
    noise_stim = Noise("Mosaic_A", 10.0, 20.0, 200.0, 0.001)
    hyper_stim = Hyperpolarizing("Mosaic_A", 0.0, 50.0)
    entries = sim.get_all_stimuli_entries()
    assert len(entries) == 2
    assert entries[0] == noise_stim
    assert entries[1] == hyper_stim


def test_condition_parameters():
    sim = SonataSimulationConfig(cond_params_conf_path)
    conditions = sim.condition_parameters()
    assert conditions == Conditions(
        mech_conditions=MechanismConditions(
            ampanmda=ConditionEntry(minis_single_vesicle=0, init_depleted=1),
            gabaab=ConditionEntry(minis_single_vesicle=None, init_depleted=None),
            glusynapse=ConditionEntry(minis_single_vesicle=None, init_depleted=None),
        ),
        celsius=34.0,
        v_init=-80.0,
        extracellular_calcium=None,
        randomize_gaba_rise_time=False,
    )


def test_connection_entries():
    sim = SonataSimulationConfig(hipp_sim_with_projection)
    entries = sim.connection_entries()
    assert len(entries) == 4
    assert entries[0] == ConnectionOverrides(
        source="Mosaic",
        target="Mosaic",
        delay=None,
        weight=1.0,
        spont_minis=0.01,
        synapse_configure=None,
        mod_override=None,
    )
    assert entries[2].synapse_configure == "%s.NMDA_ratio = 1.22 %s.tau_r_NMDA = 3.9 %s.tau_d_NMDA = 148.5"
    assert entries[-1] == ConnectionOverrides(
        source="Excitatory",
        target="Mosaic",
        delay=None,
        weight=None,
        spont_minis=None,
        synapse_configure="%s.mg = 1.0",
        mod_override=None
    )


def test_connection_override():
    sim = SonataSimulationConfig(hipp_sim_with_projection)

    entries = sim.connection_entries()
    assert len(entries) == 4

    connection_override = ConnectionOverrides(
        source="Excitatory",
        target="Mosaic",
        delay=2,
        weight=2.0,
        spont_minis=0.1,
        synapse_configure="%s.mg = 1.4",
        mod_override=None
    )
    sim.add_connection_override(connection_override)

    entries = sim.connection_entries()
    assert len(entries) == 5
    assert entries[-1] == connection_override

    # overrides are not added multiple times
    entries = sim.connection_entries()
    entries = sim.connection_entries()
    assert len(entries) == 5


def test_get_all_projection_names():
    sim_dir = cond_params_conf_path.parent
    sim_config = cond_params_conf_path.name
    with cwd(sim_dir):
        sim = SonataSimulationConfig(sim_config)
        assert sim.get_all_projection_names() == []


def test_seeds():
    sim = SonataSimulationConfig(cond_params_conf_path)
    assert sim.base_seed == 1
    assert sim.synapse_seed == 0
    assert sim.ionchannel_seed == 0
    assert sim.stimulus_seed == 0
    assert sim.minis_seed == 0


def test_rng_mode():
    sim = SonataSimulationConfig(cond_params_conf_path)
    assert sim.rng_mode == "Random123"


def test_spike_threshold():
    sim = SonataSimulationConfig(cond_params_conf_path)
    assert sim.spike_threshold == -30.0


def test_spike_location():
    sim = SonataSimulationConfig(cond_params_conf_path)
    assert sim.spike_location == "soma"


def test_duration():
    sim = SonataSimulationConfig(cond_params_conf_path)
    assert sim.duration == 50.0


def test_dt():
    sim = SonataSimulationConfig(cond_params_conf_path)
    assert sim.dt == 0.025


def test_forward_skip():
    sim = SonataSimulationConfig(cond_params_conf_path)
    assert sim.forward_skip is None


def test_celsius():
    sim = SonataSimulationConfig(cond_params_conf_path)
    assert sim.celsius == 34.0


def test_v_init():
    sim = SonataSimulationConfig(cond_params_conf_path)
    assert sim.v_init == -80.0


def test_output_root_path():
    sim = SonataSimulationConfig(cond_params_conf_path)
    assert Path(sim.output_root_path).name == "output_sonata"


def test_extracellular_calcium():
    sim = SonataSimulationConfig(cond_params_conf_path)
    assert sim.extracellular_calcium is None



"""Testing SSim with SONATA simulations."""

from pathlib import Path

import numpy as np
import pytest

from bluecellulab import SSim

parent_dir = Path(__file__).resolve().parent


INPUTS_TO_TEST = ["noinput", "hypamp", "ornstein", "shotnoise"]


@pytest.mark.v6
@pytest.mark.parametrize("input_type", ["noinput", "hypamp"])
def test_sim_quick_scx_sonata(input_type):
    """Test against sim results of quick_scx_sonata."""
    # Path to the SONATA simulation
    sonata_sim_path = (
        parent_dir
        / "examples"
        / "sim_quick_scx_sonata"
        / f"simulation_config_{input_type}.json"
    )

    # Create SSim object
    sim = SSim(sonata_sim_path)
    cell_id = ("NodeA", 2)  # has a spike + hyperpolarises
    sim.instantiate_gids(cell_id, add_stimuli=True)
    t_stop = 10.0
    sim.run(t_stop)

    # Get the voltage trace
    time = sim.get_time_trace(1)
    voltage = sim.get_voltage_trace(cell_id, 0, t_stop, 1)
    voltage = voltage[:len(voltage) - 1]  # remove last point, mainsim produces 1 less
    time = time[:len(time) - 1]  # remove last point, mainsim produces 1 less
    mainsim_voltage = sim.get_mainsim_voltage_trace(cell_id)
    voltage_diff = voltage - mainsim_voltage
    rms_error = np.sqrt(np.mean(voltage_diff ** 2))
    assert rms_error < 1e-4


@pytest.mark.v6
@pytest.mark.parametrize("input_type", INPUTS_TO_TEST)
def test_sim_quick_scx_sonata_multicircuit(input_type):
    """Sonata config multicircuit test.
       Applies the stimulus defined in INPUTS_TO_TEST one by one.
    """
    sonata_sim_path = (
        parent_dir
        / "examples"
        / "sim_quick_scx_sonata_multicircuit"
        / f"simulation_config_{input_type}.json"
    )

    cell_ids = [("NodeA", 1), ("NodeA", 2)]
    # investivate NodeA, 0 further. It shows small discrepancies,
    # even on a single population circuit
    # Create SSim object
    sim = SSim(sonata_sim_path)
    sim.instantiate_gids(cell_ids, add_stimuli=True)
    t_stop = 20.0
    sim.run(t_stop)
    for cell_id in cell_ids:
        voltage = sim.get_voltage_trace(cell_id, 0, t_stop, 1)
        voltage = voltage[:len(voltage) - 1]  # remove last point, mainsim produces 1 less
        mainsim_voltage = sim.get_mainsim_voltage_trace(cell_id)
        voltage_diff = voltage - mainsim_voltage
        rms_error = np.sqrt(np.mean(voltage_diff ** 2))
        assert rms_error < 0.25


@pytest.mark.v6
def test_ssim_intersect_pre_gids_multipopulation():
    """Test instantiate_gids with intersect_pre_gids on Sonata."""
    sonata_sim_path = (
        parent_dir
        / "examples"
        / "sim_quick_scx_sonata_multicircuit"
        / "simulation_config_noinput.json"
    )
    cell_ids = [("NodeA", 0), ("NodeA", 1)]

    sim = SSim(sonata_sim_path)
    sim.instantiate_gids(cell_ids, add_synapses=True)
    assert len([x.synapses for x in sim.cells.values()][0]) == 6
    assert len([x.synapses for x in sim.cells.values()][1]) == 2

    # pre gids are intersected, synapses are filtered
    sim2 = SSim(sonata_sim_path)
    sim2.instantiate_gids(cell_ids, add_synapses=True, intersect_pre_gids=[("NodeB", 0)])
    assert len([x.synapses for x in sim2.cells.values()][0]) == 2
    assert len([x.synapses for x in sim2.cells.values()][1]) == 0


@pytest.mark.v6
def test_merge_pre_spike_trains_edge_case():
    """Test to make sure connections get initialised.
   for fixing empty cell_info_dict['connections'] bug."""
    sonata_sim_path = (
        parent_dir
        / "examples"
        / "sim_quick_scx_sonata_multicircuit"
        / "simulation_config_noinput.json"
    )
    cell_id = ("NodeA", 0)
    ssim = SSim(sonata_sim_path)
    ssim.instantiate_gids(cell_id, add_minis=True, add_replay=True,
                          add_stimuli=False, add_synapses=True,
                          intersect_pre_gids=None)
    cell_info_dict = ssim.cells[cell_id].info_dict
    assert cell_info_dict["connections"] != {}



from contextlib import contextmanager
import os


@contextmanager
def cwd(path):
    """Cwd function that can be used in a context manager."""
    old_dir = os.getcwd()
    os.chdir(path)
    yield
    os.chdir(old_dir)



"""Unit tests for node_id.py"""
from bluecellulab.circuit.node_id import create_cell_id, CellId


def test_create_cell_id():
    """SSim: Testing creation of CellId."""
    assert create_cell_id(1) == CellId("", 1)
    assert create_cell_id(("", 2)) == CellId("", 2)
    assert create_cell_id(("pop1", 2)) == CellId("pop1", 2)


def test_create_cell_ids():
    """SSim: Testing creation of CellIds."""
    from bluecellulab.circuit.node_id import create_cell_ids, CellId

    assert create_cell_ids([1, 2]) == [CellId("", 1), CellId("", 2)]
    assert create_cell_ids([("", 2), ("pop1", 2)]) == [
        CellId("", 2),
        CellId("pop1", 2),
    ]



"""Unit tests for the neuron interpreter."""

import os

import neuron
from pytest import raises
import pytest

import bluecellulab
from bluecellulab.exceptions import NeuronEvalError
from bluecellulab.importer import _load_hoc_and_mod_files
from bluecellulab.neuron_interpreter import eval_neuron

script_dir = os.path.dirname(__file__)


def test_eval_neuron():
    """Unit test for the eval_neuron function."""
    _load_hoc_and_mod_files()
    eval_neuron("neuron.h.nil", neuron=neuron)
    with raises(NeuronEvalError):
        eval_neuron("1+1")
    with raises(NeuronEvalError):
        eval_neuron("neuron.h.nil; 2-1", neuron=neuron)
    with raises(NeuronEvalError):
        eval_neuron("a=1")


@pytest.mark.v5
def test_eval_neuron_with_cell():
    """Test the eval neuron function using a cell."""
    cell = bluecellulab.Cell(
        f"{script_dir}/examples/cell_example1/test_cell.hoc",
        f"{script_dir}/examples/cell_example1",
    )

    eval_neuron("self.axonal[1](0.5)._ref_v", self=cell)
    AXON_LOC = "self.axonal[1](0.5)._ref_v"
    cell.add_recordings(["neuron.h._ref_t", AXON_LOC], dt=cell.record_dt)



"""
Synapse test file. Uses ball-and-stick models
Models simulated in N disticnt ways, all shoudl
"""


import os

import neuron
import numpy as np

import bluecellulab
from bluecellulab import create_ball_stick

script_dir = os.path.dirname(__file__)


class Params:

    """Simulation parameters"""

    def __init__(self):
        self.T_STOP = 200
        self.V_INIT = -75
        self.DT = 0.025

        self.SYN_DECAY = 2.0
        self.SYN_E = 0
        self.SYN_ACTIVATION_T = 50
        self.SYN_DELAY = 0.025
        self.SYN_G = 0.001
        self.SYN_LOC = 1.0

        self.templatefile = \
            "%s/examples/ballstick_cell/ballstick.hoc" % script_dir
        self.morphfile = \
            "%s/examples/ballstick_cell/ballstick.asc" % script_dir

        cell = bluecellulab.Cell(self.templatefile, self.morphfile)
        self.soma_L, self.soma_D, self.soma_A = \
            cell.soma.L, cell.soma.diam, neuron.h.area(
                0.5, sec=cell.soma)
        # print 'SOMA L=%f, diam=%f,surf=%f' % (self.soma_L, self.soma_D,
        # self.soma_A)

        self.dend0_L, self.dend0_D, self.dend0_A = cell.basal[0].L, cell.basal[
            0].diam, neuron.h.area(
            0.5, sec=cell.basal[0])
        self.dend0_NSEG = cell.basal[0].nseg

        # print 'DENDRITE L=%f, diam=%f,surf=%f' % (self.dend0_L, self.dend0_D,
        # self.dend0_A)

        # I assume uniform passive properties shared by the soma and dendrites
        self.CM = cell.soma.cm
        self.RM = 1.0 / cell.soma(0.5).g_pas
        self.RA = cell.soma.Ra
        self.EL = cell.soma(0.5).e_pas

        self.d_derived = 10
        self.l_derived = self.soma_A / (2 * np.pi * self.d_derived / 2.0)


def run_pyneuron(soma_l, soma_d, params):
    """Run ballstick with PyNeuron"""
    soma = neuron.h.Section()
    soma.L = soma_l
    soma.diam = soma_d
    soma.nseg = 1
    soma.cm = params.CM
    soma.Ra = params.RA

    soma.insert('pas')
    soma(0.5).e_pas = params.EL
    soma(0.5).g_pas = 1.0 / params.RM

    dend = neuron.h.Section()
    dend.L = params.dend0_L
    dend.diam = params.dend0_D
    dend.nseg = params.dend0_NSEG

    dend.cm = params.CM
    dend.Ra = params.RA
    dend.insert('pas')
    for seg in dend:
        seg.e_pas = params.EL
        seg.g_pas = 1.0 / params.RM
    dend.connect(soma, 0.5, 0)  # mid-soma to begin-den

    syn = neuron.h.ExpSyn(params.SYN_LOC, sec=dend)
    syn.tau = params.SYN_DECAY
    syn.e = params.SYN_E

    ns = neuron.h.NetStim()
    ns.interval = 100000
    ns.number = 1
    ns.start = params.SYN_ACTIVATION_T
    ns.noise = 0

    nc = neuron.h.NetCon(ns, syn, 0, params.SYN_DELAY, params.SYN_G)

    v_vec = neuron.h.Vector()
    t_vec = neuron.h.Vector()
    v_vec.record(soma(0.5)._ref_v)
    t_vec.record(neuron.h._ref_t)

    neuron.h.finitialize(params.V_INIT)
    neuron.h.dt = params.DT
    neuron.run(params.T_STOP)

    voltage = np.array(v_vec)
    time = np.array(t_vec)

    del syn
    del ns
    del nc

    return time, voltage


def run_pyneuron_with_template(params):
    """Run ballstick with PyNeuron and template"""

    neuron.h.load_file(params.templatefile)
    cell = neuron.h.ballstick_cell(0, params.morphfile)
    basal = cell.getCell().dend[0]
    soma = cell.getCell().soma[0]
    syn = neuron.h.ExpSyn(params.SYN_LOC, sec=basal)
    syn.tau = params.SYN_DECAY
    syn.e = params.SYN_E
    ns = neuron.h.NetStim()
    ns.interval = 100000
    ns.number = 1
    ns.start = params.SYN_ACTIVATION_T
    ns.noise = 0
    nc = neuron.h.NetCon(ns, syn, 0, params.SYN_DELAY, params.SYN_G)

    v_vec = neuron.h.Vector()
    t_vec = neuron.h.Vector()
    v_vec.record(soma(0.5)._ref_v)
    t_vec.record(neuron.h._ref_t)

    neuron.h.finitialize(params.V_INIT)
    neuron.h.dt = params.DT
    neuron.run(params.T_STOP)

    voltage = np.array(v_vec)
    time = np.array(t_vec)

    cell.getCell().clear()

    del syn
    del ns
    del nc
    del cell

    return time, voltage


def run_bluecellulab(params):
    """Run ballstick with bluecellulab"""
    cell = bluecellulab.Cell(params.templatefile, params.morphfile)
    syn = neuron.h.ExpSyn(params.SYN_LOC, sec=cell.basal[0])
    syn.tau = params.SYN_DECAY
    syn.e = params.SYN_E
    ns = neuron.h.NetStim()
    ns.interval = 100000
    ns.number = 1
    ns.start = params.SYN_ACTIVATION_T
    ns.noise = 0
    nc = neuron.h.NetCon(ns, syn, 0, params.SYN_DELAY, params.SYN_G)

    sim = bluecellulab.Simulation()
    sim.add_cell(cell)
    sim.run(params.T_STOP, v_init=params.V_INIT, cvode=False, dt=params.DT)
    bluecellulab_t = cell.get_time()
    bluecellulab_v = cell.get_soma_voltage()

    del sim
    del syn
    del ns
    del nc
    del cell

    return bluecellulab_t, bluecellulab_v


def run_analytic(params):
    ''' write config file for Willem '''
    f_name = '%s/examples/ballstick_cell/bs.cfg' % script_dir
    outF = open(f_name, 'w')
    outF.write('[neuron]\n')
    outF.write('CM: %f\n' % (params.CM))
    outF.write('RM: %f\n' % (params.RM))
    outF.write('RA: %f\n' % (params.RA))
    outF.write('EL: %f\n\n' % (params.EL))
    outF.write('[soma]\n')
    outF.write('D: ' + str(params.d_derived) + '\n')
    outF.write('L: ' + str(params.l_derived) + '\n\n')
    outF.write('[morph]\n')
    outF.write('lengths: [%f]\n' % (params.dend0_L))
    outF.write('diams: [%f]\n' % (params.dend0_D))
    outF.close()

    v_willem, t_willem = gfc.compute_system(
        [[params.SYN_ACTIVATION_T]], params.T_STOP, params.DT,
        conffile_name=f_name, numsyn=1, syndend=[0], synloc=[params.SYN_LOC],
        gbar=[params.SYN_G], decay=[params.SYN_DECAY], E_rev=[params.SYN_E])
    return t_willem, v_willem


def test_expsyn_pyneuron_vs_bluecellulab(graph=False):
    """Ballstick: Test ballstick with expsyn between pyneuron and bluecellulab"""

    """
    The real stuff, part I
    Run the ball-and-stick model by 1. PyNEURON, 2. bluecellulab, 3. analytic
    """

    params = Params()  # define all the parameters

    # Run in pure PyNeuron, with using the template
    pyneuron_t, pyneuron_v = run_pyneuron(
        params.l_derived, params.d_derived, params)

    # Run with template in PyNeuron
    pyneuron_template_t, pyneuron_template_v = run_pyneuron_with_template(
        params)

    # Run with bluecellulab
    bluecellulab_t, bluecellulab_v = run_bluecellulab(params)

    # Run with analytic solution
    analytic_expsyn_path = \
        "%s/examples/ballstick_cell/analytic_expsyn.txt" % script_dir

    analytic_t, analytic_v = np.loadtxt(analytic_expsyn_path)

    assert len(analytic_v) == 8000
    pyneuron_rms_error = np.sqrt(
        np.mean(
            (analytic_v -
             pyneuron_v[
                 :len(analytic_v)]) ** 2))
    pyneuron_template_rms_error = np.sqrt(
        np.mean(
            (analytic_v -
             pyneuron_template_v[
                 :len(analytic_v)]) ** 2))
    bluecellulab_rms_error = np.sqrt(
        np.mean(
            (analytic_v -
             bluecellulab_v[
                 :len(analytic_v)]) ** 2))
    assert pyneuron_rms_error < 0.1
    assert pyneuron_template_rms_error < 0.1
    assert bluecellulab_rms_error < 0.1

    if graph:
        import pylab
        pylab.plot(bluecellulab_t, bluecellulab_v, 'g', label='bluecellulab', alpha=0.5)
        pylab.plot(pyneuron_t, pyneuron_v, 'b', label='PyNeuron', linestyle='--')
        pylab.plot(
            pyneuron_template_t,
            pyneuron_template_v,
            'r',
            label='PyNeuron with template',
            linestyle='-.',
            alpha=0.5)
        pylab.plot(analytic_t, analytic_v, '-', label='Analytic', linestyle=":", alpha=0.5)
        pylab.legend(loc=0)
        pylab.show()


def test_ballstick_load():
    """Ballstick: Test if dimensions of ballstick load correctly"""
    params = Params()

    cell = bluecellulab.Cell(params.templatefile, params.morphfile)
    assert abs(cell.soma.L - 19.6) < 0.001
    assert abs(cell.soma.diam - 10.229) < 0.001
    assert abs(neuron.h.area(0.5, sec=cell.soma) - 872.567) < 0.001
    assert abs(cell.basal[0].L - 200.0) < 0.001
    assert abs(cell.basal[0].diam - 3.0) < 0.001
    assert abs(neuron.h.area(0.5, sec=cell.basal[0]) - 9.424) < 0.001

    del cell


def test_create_ballstick():
    """Unit test for cell/ballstick."""
    cell = create_ball_stick()
    assert isinstance(cell, bluecellulab.Cell)
    assert cell.cell.gid == 0
    assert cell.cell.getHypAmp() == -0.070557


def main():
    """main"""
    test_expsyn_pyneuron_vs_bluecellulab(graph=True)


if __name__ == "__main__":
    main()



# Copyright 2012-2024 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""RNG settings of bluecellulab."""

import logging

from typing import Optional

import neuron
from bluecellulab.circuit.config.definition import SimulationConfig
from bluecellulab.exceptions import UndefinedRNGException
from bluecellulab.importer import load_hoc_and_mod_files

logger = logging.getLogger(__name__)


class RNGSettings:
    """Singleton object that represents RNG settings in bluecellulab."""

    _instance = None

    @classmethod
    def get_instance(cls):
        """Return the instance of the class."""
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    @load_hoc_and_mod_files
    def __init__(
        self,
        mode="Random123",
        base_seed=0,
        synapse_seed=0,
        ionchannel_seed=0,
        stimulus_seed=0,
        minis_seed=0,
    ) -> None:
        self.mode = mode
        self.base_seed = base_seed
        self.synapse_seed = synapse_seed
        self.ionchannel_seed = ionchannel_seed
        self.stimulus_seed = stimulus_seed
        self.minis_seed = minis_seed

    def set_seeds(
            self,
            mode: Optional[str] = None,
            sim_config: Optional[SimulationConfig] = None,
            base_seed: Optional[int] = None):
        """Constructor.

        Parameters
        ----------
        mode : rng mode, if not specified mode is taken from circuit_access
        sim_config: simulation config object, if present seeds are read from simulation
        base_seed: base seed for entire sim, overrides config value
        """
        self._mode = ""
        if mode is None:
            if sim_config is not None:
                self.mode = sim_config.rng_mode if sim_config else "Compatibility"
            else:
                self.mode = "Random123"
        else:
            self.mode = mode

        logger.debug("Setting rng mode to: %s", self._mode)

        if base_seed is None:
            self.base_seed = sim_config.base_seed if sim_config else 0
        else:
            self.base_seed = base_seed
        neuron.h.globalSeed = self.base_seed

        if self._mode == 'Random123':
            rng = neuron.h.Random()
            rng.Random123_globalindex(self.base_seed)

        if sim_config:
            self.synapse_seed = sim_config.synapse_seed
            self.ionchannel_seed = sim_config.ionchannel_seed
            self.stimulus_seed = sim_config.stimulus_seed
            self.minis_seed = sim_config.minis_seed

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, new_val):
        """Setter method for the mode."""

        options = {"Compatibility": 0, "Random123": 1, "UpdatedMCell": 2}
        if new_val not in options:
            raise UndefinedRNGException(
                "SSim: RNG mode %s not in accepted list: %s"
                % (self.mode, list(options.keys()))
            )
        else:
            neuron.h.rngMode = options[new_val]
            self._mode = new_val

    def __repr__(self) -> str:
        """Returns a string representation of the object."""
        return "RNGSettings(mode={mode}, base_seed={base_seed}, " \
               "synapse_seed={synapse_seed}, " \
               "ionchannel_seed={ionchannel_seed}, " \
               "stimulus_seed={stimulus_seed}, " \
               "minis_seed={minis_seed})".format(
                   mode=self.mode,
                   base_seed=self.base_seed,
                   synapse_seed=self.synapse_seed,
                   ionchannel_seed=self.ionchannel_seed,
                   stimulus_seed=self.stimulus_seed,
                   minis_seed=self.minis_seed)



# Copyright 2012-2024 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for calculating certain properties of Neurons."""


from __future__ import annotations
from pathlib import Path
from typing import Optional, Tuple
import logging

import neuron
import numpy as np

import bluecellulab
from bluecellulab.circuit.circuit_access import EmodelProperties
from bluecellulab.exceptions import UnsteadyCellError
from bluecellulab.utils import CaptureOutput, IsolatedProcess

logger = logging.getLogger(__name__)


def calculate_input_resistance(
    template_path: str | Path,
    morphology_path: str | Path,
    template_format: str,
    emodel_properties: EmodelProperties | None,
    current_delta: float = 0.01,
) -> float:
    """Calculate the input resistance at rest of the cell."""
    rest_voltage = calculate_SS_voltage(
        template_path, morphology_path, template_format, emodel_properties, 0.0
    )
    step_voltage = calculate_SS_voltage(
        template_path,
        morphology_path,
        template_format,
        emodel_properties,
        current_delta,
    )

    voltage_delta = step_voltage - rest_voltage

    return voltage_delta / current_delta


def calculate_SS_voltage(
    template_path: str | Path,
    morphology_path: str | Path,
    template_format: str,
    emodel_properties: EmodelProperties | None,
    step_level: float,
) -> float:
    """Calculate the steady state voltage at a certain current step."""
    with IsolatedProcess() as runner:
        SS_voltage = runner.apply(
            calculate_SS_voltage_subprocess,
            [
                template_path,
                morphology_path,
                template_format,
                emodel_properties,
                step_level,
            ],
        )
    return SS_voltage


def calculate_SS_voltage_subprocess(
    template_path: str | Path,
    morphology_path: str | Path,
    template_format: str,
    emodel_properties: EmodelProperties | None,
    step_level: float,
    check_for_spiking=False,
    spike_threshold=-20.0,
) -> float:
    """Subprocess wrapper of calculate_SS_voltage.

    This code should be run in a separate process. If check_for_spiking
    is True, this function will return None if the cell spikes from
    100ms to the end of the simulation indicating no steady state was
    reached.
    """
    cell = bluecellulab.Cell(
        template_path=template_path,
        morphology_path=morphology_path,
        template_format=template_format,
        emodel_properties=emodel_properties,
    )
    cell.add_ramp(500, 5000, step_level, step_level)
    simulation = bluecellulab.Simulation()
    simulation.run(1000, cvode=template_accepts_cvode(template_path))
    time = cell.get_time()
    voltage = cell.get_soma_voltage()
    SS_voltage = np.mean(voltage[np.where((time < 1000) & (time > 800))])
    cell.delete()

    if check_for_spiking:
        # check for voltage crossings
        if len(np.nonzero(voltage[np.where(time > 100.0)] > spike_threshold)[0]) > 0:
            raise UnsteadyCellError(
                "Cell spikes from 100ms to the end of the simulation."
            )

    return SS_voltage


def holding_current_subprocess(v_hold, enable_ttx, cell_kwargs):
    """Subprocess wrapper of holding_current."""
    cell = bluecellulab.Cell(**cell_kwargs)

    if enable_ttx:
        cell.enable_ttx()

    vclamp = neuron.h.SEClamp(0.5, sec=cell.soma)
    vclamp.rs = 0.01
    vclamp.dur1 = 2000
    vclamp.amp1 = v_hold

    simulation = bluecellulab.Simulation()
    simulation.run(1000, cvode=False)

    i_hold = vclamp.i
    v_control = vclamp.vc

    cell.delete()

    return i_hold, v_control


def holding_current(
    v_hold: float,
    cell_id: int | tuple[str, int],
    circuit_path: str | Path,
    enable_ttx=False,
) -> Tuple[float, float]:
    """Calculate the holding current necessary for a given holding voltage."""
    cell_id = bluecellulab.circuit.node_id.create_cell_id(cell_id)
    ssim = bluecellulab.SSim(circuit_path)

    cell_kwargs = ssim.fetch_cell_kwargs(cell_id)
    with IsolatedProcess() as runner:
        i_hold, v_control = runner.apply(
            holding_current_subprocess, [v_hold, enable_ttx, cell_kwargs]
        )

    return i_hold, v_control


def template_accepts_cvode(template_name: str | Path) -> bool:
    """Return True if template_name can be run with cvode."""
    with open(template_name, "r") as template_file:
        template_content = template_file.read()
    if "StochKv" in template_content:
        accepts_cvode = False
    else:
        accepts_cvode = True
    return accepts_cvode


def search_hyp_current(
    template_path: str | Path,
    morphology_path: str | Path,
    template_format: str,
    emodel_properties: Optional[EmodelProperties],
    target_voltage: float,
    min_current: float,
    max_current: float,
) -> float:
    """Search current necessary to bring cell to -85 mV."""
    med_current = min_current + abs(min_current - max_current) / 2
    new_target_voltage = calculate_SS_voltage(
        template_path,
        morphology_path,
        template_format,
        emodel_properties,
        med_current,
    )
    logger.info("Detected voltage: %f" % new_target_voltage)
    if abs(new_target_voltage - target_voltage) < 0.5:
        return med_current
    elif new_target_voltage > target_voltage:
        return search_hyp_current(
            template_path=template_path,
            morphology_path=morphology_path,
            template_format=template_format,
            emodel_properties=emodel_properties,
            target_voltage=target_voltage,
            min_current=min_current,
            max_current=med_current,
        )
    else:  # new_target_voltage < target_voltage:
        return search_hyp_current(
            template_path=template_path,
            morphology_path=morphology_path,
            template_format=template_format,
            emodel_properties=emodel_properties,
            target_voltage=target_voltage,
            min_current=med_current,
            max_current=max_current,
        )


def detect_hyp_current(
    template_path: str | Path,
    morphology_path: str | Path,
    template_format: str,
    emodel_properties: EmodelProperties | None,
    target_voltage: float,
) -> float:
    """Search current necessary to bring cell to -85 mV.

    Compared to using NEURON's SEClamp object, the binary search better
    replicates what experimentalists use
    """
    return search_hyp_current(
        template_path=template_path,
        morphology_path=morphology_path,
        template_format=template_format,
        emodel_properties=emodel_properties,
        target_voltage=target_voltage,
        min_current=-1.0,
        max_current=0.0,
    )


def detect_spike_step(
    template_path: str | Path,
    morphology_path: str | Path,
    template_format: str,
    emodel_properties: EmodelProperties | None,
    hyp_level: float,
    inj_start: float,
    inj_stop: float,
    step_level: float,
) -> bool:
    """Detect if there is a spike at a certain step level."""
    with IsolatedProcess() as runner:
        spike_detected = runner.apply(
            detect_spike_step_subprocess,
            [
                template_path,
                morphology_path,
                template_format,
                emodel_properties,
                hyp_level,
                inj_start,
                inj_stop,
                step_level,
            ],
        )
    return spike_detected


def detect_spike_step_subprocess(
    template_path: str | Path,
    morphology_path: str | Path,
    template_format: str,
    emodel_properties: EmodelProperties | None,
    hyp_level: float,
    inj_start: float,
    inj_stop: float,
    step_level: float
) -> bool:
    """Detect if there is a spike at a certain step level."""
    cell = bluecellulab.Cell(
        template_path=template_path,
        morphology_path=morphology_path,
        template_format=template_format,
        emodel_properties=emodel_properties)
    cell.add_ramp(0, 5000, hyp_level, hyp_level)
    cell.add_ramp(inj_start, inj_stop, step_level, step_level)
    simulation = bluecellulab.Simulation()
    simulation.run(int(inj_stop), cvode=template_accepts_cvode(template_path))

    time = cell.get_time()
    voltage = cell.get_soma_voltage()
    time_step = time[np.where((time > inj_start) & (time < inj_stop))]
    voltage_step = voltage[np.where((time_step > inj_start) & (time_step < inj_stop))]
    spike_detected = detect_spike(voltage_step)

    cell.delete()

    return spike_detected


def detect_spike(voltage: np.ndarray) -> bool:
    """Detect if there is a spike in the voltage trace."""
    if len(voltage) == 0:
        return False
    else:
        return bool(np.max(voltage) > -20)  # bool not np.bool_


def search_threshold_current(
    template_name: str | Path,
    morphology_path: str | Path,
    template_format: str,
    emodel_properties: EmodelProperties | None,
    hyp_level: float,
    inj_start: float,
    inj_stop: float,
    min_current: float,
    max_current: float,
):
    """Search current necessary to reach threshold."""
    med_current = min_current + abs(min_current - max_current) / 2
    logger.info("Med current %d" % med_current)

    spike_detected = detect_spike_step(
        template_name, morphology_path, template_format, emodel_properties,
        hyp_level, inj_start, inj_stop, med_current
    )
    logger.info("Spike threshold detection at: %f nA" % med_current)

    if abs(max_current - min_current) < .01:
        return max_current
    elif spike_detected:
        return search_threshold_current(template_name, morphology_path,
                                        template_format, emodel_properties,
                                        hyp_level, inj_start, inj_stop,
                                        min_current, med_current)
    else:
        return search_threshold_current(template_name, morphology_path,
                                        template_format, emodel_properties,
                                        hyp_level, inj_start, inj_stop,
                                        med_current, max_current)


def check_empty_topology() -> bool:
    """Return true if NEURON simulator topology command is empty."""
    with CaptureOutput() as stdout:
        neuron.h.topology()

    return stdout == ['', '']



# Copyright 2012-2024 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for handling NEURON hoc templates."""

from __future__ import annotations

import hashlib
import os
from pathlib import Path
import re
import string
from typing import Optional

import neuron

from bluecellulab.circuit import EmodelProperties
from bluecellulab.exceptions import BluecellulabError
from bluecellulab.type_aliases import HocObjectType

import logging

logger = logging.getLogger(__name__)


def public_hoc_cell(cell: HocObjectType) -> HocObjectType:
    """Retrieve the hoc cell to access public hoc functions/attributes."""
    if hasattr(cell, "getCell"):
        return cell.getCell()
    elif hasattr(cell, "CellRef"):
        return cell.CellRef
    else:
        raise BluecellulabError("""Public cell properties cannot be accessed
         from the hoc model. Either getCell() or CellRef needs to be provided""")


class NeuronTemplate:
    """NeuronTemplate representation."""

    def __init__(
        self, template_filepath: str | Path, morph_filepath: str | Path,
        template_format: str, emodel_properties: Optional[EmodelProperties]
    ) -> None:
        """Load the hoc template and init object."""
        if isinstance(template_filepath, Path):
            template_filepath = str(template_filepath)
        if isinstance(morph_filepath, Path):
            morph_filepath = str(morph_filepath)

        if not os.path.exists(template_filepath):
            raise FileNotFoundError(f"Couldn't find template file: {template_filepath}")
        if not os.path.exists(morph_filepath):
            raise FileNotFoundError(f"Couldn't find morphology file: {morph_filepath}")

        self.template_name = self.load(template_filepath)
        self.morph_filepath = morph_filepath
        self.template_format = template_format
        self.emodel_properties = emodel_properties

    def get_cell(self, gid: Optional[int]) -> HocObjectType:
        """Returns the hoc object matching the template format."""
        morph_dir, morph_fname = os.path.split(self.morph_filepath)
        if self.template_format == "v6":
            attr_names = getattr(
                neuron.h, self.template_name.split('_bluecellulab')[0] + "_NeededAttributes", None
            )
            if attr_names is not None:
                if self.emodel_properties is None:
                    raise BluecellulabError(
                        "EmodelProperties must be provided for template "
                        "format v6 that specifies _NeededAttributes"
                    )
                cell = getattr(neuron.h, self.template_name)(
                    gid,
                    morph_dir,
                    morph_fname,
                    *[self.emodel_properties.__getattribute__(name) for name in attr_names.split(";")]
                )
            else:
                cell = getattr(neuron.h, self.template_name)(
                    gid,
                    morph_dir,
                    morph_fname,
                )
        elif self.template_format == "bluepyopt":
            cell = getattr(neuron.h, self.template_name)(morph_dir, morph_fname)
        else:
            cell = getattr(neuron.h, self.template_name)(gid, self.morph_filepath)

        return cell

    def load(self, template_filename: str) -> str:
        """Read a cell template. If template name already exists, rename it.

        Args:
            template_filename: path string containing template file.

        Returns:
            resulting template name
        """
        with open(template_filename) as template_file:
            template_content = template_file.read()

        match = re.search(r"begintemplate\s*(\S*)", template_content)
        template_name = match.group(1)  # type:ignore

        logger.debug("This Neuron version supports renaming templates, enabling...")
        # add bluecellulab to the template name, so that we don't interfere with
        # templates load outside of bluecellulab
        template_name = "%s_bluecellulab" % template_name
        template_name = get_neuron_compliant_template_name(template_name)
        obj_address = hex(id(self))
        template_name = f"{template_name}_{obj_address}"

        template_content = re.sub(
            r"begintemplate\s*(\S*)",
            "begintemplate %s" % template_name,
            template_content,
        )
        template_content = re.sub(
            r"endtemplate\s*(\S*)",
            "endtemplate %s" % template_name,
            template_content,
        )

        neuron.h(template_content)

        return template_name


def shorten_and_hash_string(label: str, keep_length=40, hash_length=9) -> str:
    """Converts a string to a shorter string if required.

    Args:
        label: A string to be converted.
        keep_length: Length of the original string to keep.
        hash_length: Length of the hash to generate, should not be more than 20.

    Returns:
        If the length of the original label is shorter than the sum of 'keep_length'
        and 'hash_length' plus one, the original string is returned. Otherwise, a
        string with structure <partial>_<hash> is returned, where <partial> is the
        first part of the original string with length equal to <keep_length> and the
        last part is a hash of 'hash_length' characters, based on the original string.
    """
    if hash_length > 20:
        raise ValueError(
            "Parameter hash_length should not exceed 20, "
            " received: {}".format(hash_length)
        )

    if len(label) <= keep_length + hash_length + 1:
        return label

    hash_string = hashlib.sha1(label.encode("utf-8")).hexdigest()
    return "{}_{}".format(label[0:keep_length], hash_string[0:hash_length])


def check_compliance_with_neuron(template_name: str) -> bool:
    """Verify that a given name is compliant with the rules for a NEURON.

    A name should be a non-empty alphanumeric string, and start with a
    letter. Underscores are allowed. The length should not exceed 50
    characters.
    """
    max_len = 50
    return (
        bool(template_name)
        and template_name[0].isalpha()
        and template_name.replace("_", "").isalnum()
        and len(template_name) <= max_len
    )


def get_neuron_compliant_template_name(name: str) -> str:
    """Get template name that is compliant with NEURON based on given name."""
    template_name = name
    if not check_compliance_with_neuron(template_name):
        template_name = template_name.lstrip(string.digits).replace("-", "_")
        template_name = shorten_and_hash_string(
            template_name, keep_length=40, hash_length=9
        )
        logger.debug("Converted template name %s to %s to make it "
                     "NEURON compliant" % (name, template_name))
    return template_name



"""Test for the verbosity module."""


import logging
import pytest


@pytest.mark.unit
def test_verbose_env():
    """bluecellulab: Test env verbose level var"""
    import os
    os.environ['BLUECELLULAB_VERBOSE_LEVEL'] = "10"

    import bluecellulab  # NOQA

    bluecellulab.set_verbose_from_env()

    assert bluecellulab.ENV_VERBOSE_LEVEL == "10"
    assert bluecellulab.VERBOSE_LEVEL == 10

    del os.environ['BLUECELLULAB_VERBOSE_LEVEL']

    bluecellulab.set_verbose_from_env()
    assert bluecellulab.ENV_VERBOSE_LEVEL is None
    assert bluecellulab.VERBOSE_LEVEL == 10
    assert logging.getLogger('bluecellulab').getEffectiveLevel() == logging.DEBUG

    bluecellulab.set_verbose(0)
    assert bluecellulab.VERBOSE_LEVEL == 0
    assert logging.getLogger('bluecellulab').getEffectiveLevel() == logging.CRITICAL

    bluecellulab.set_verbose(1)
    assert bluecellulab.VERBOSE_LEVEL == 1
    assert logging.getLogger('bluecellulab').getEffectiveLevel() == logging.ERROR

    bluecellulab.set_verbose(2)
    assert bluecellulab.VERBOSE_LEVEL == 2
    assert logging.getLogger('bluecellulab').getEffectiveLevel() == logging.WARNING

    bluecellulab.set_verbose(5)
    assert bluecellulab.VERBOSE_LEVEL == 5
    assert logging.getLogger('bluecellulab').getEffectiveLevel() == logging.INFO



from unittest import mock

import pytest
import neuron

from bluecellulab.circuit.config.sections import ConditionEntry, Conditions, MechanismConditions
from bluecellulab.importer import _load_hoc_and_mod_files
from bluecellulab.simulation.neuron_globals import set_global_condition_parameters, set_init_depleted_values, set_minis_single_vesicle_values, set_tstop_value


def test_set_tstop_value():
    _load_hoc_and_mod_files()  # this is a run_once function
    set_tstop_value(100.0)
    assert neuron.h.tstop == 100.0


@mock.patch("neuron.h")
def test_set_global_condition_parameters_no_mechs(mocked_h):
    del mocked_h.cao_CR_GluSynapse  # delete the attribute for the mock
    calcium_condition = Conditions(extracellular_calcium=2.0)
    set_global_condition_parameters(calcium_condition)
    assert not hasattr(neuron.h, "cao_CR_GluSynapse")

    mech_conditions = Conditions(mech_conditions=MechanismConditions(None, None, None))
    set_global_condition_parameters(mech_conditions)


@pytest.mark.v6
def test_set_global_condition_parameters():
    calcium_condition = Conditions(extracellular_calcium=2.015)
    set_global_condition_parameters(calcium_condition)
    assert neuron.h.cao_CR_GluSynapse == 2.015


@pytest.mark.v6
def test_set_init_depleted_values():
    mech_conditions = MechanismConditions(
        ampanmda=ConditionEntry(minis_single_vesicle=None, init_depleted=True),
        gabaab=ConditionEntry(minis_single_vesicle=None, init_depleted=False),
        glusynapse=ConditionEntry(minis_single_vesicle=None, init_depleted=None),
    )
    set_init_depleted_values(mech_conditions)
    assert neuron.h.init_depleted_ProbAMPANMDA_EMS == 1.0
    assert neuron.h.init_depleted_ProbGABAAB_EMS == 0.0
    assert neuron.h.init_depleted_GluSynapse == 0.0


@pytest.mark.v6
def test_set_minis_single_vesicle_values():
    mech_conditions = MechanismConditions(
        ampanmda=ConditionEntry(minis_single_vesicle=True, init_depleted=None),
        gabaab=ConditionEntry(minis_single_vesicle=False, init_depleted=None),
        glusynapse=ConditionEntry(minis_single_vesicle=None, init_depleted=None),
    )
    set_minis_single_vesicle_values(mech_conditions)
    assert neuron.h.minis_single_vesicle_ProbAMPANMDA_EMS == 1.0
    assert neuron.h.minis_single_vesicle_ProbGABAAB_EMS == 0.0
    assert neuron.h.minis_single_vesicle_GluSynapse == 0.0



"""Unit tests for the RNGSettings class."""

import neuron

from bluecellulab import RNGSettings
from bluecellulab.exceptions import UndefinedRNGException


def test_setting_rngmodes():
    """Test the setting of rng mode."""
    rng_obj = RNGSettings.get_instance()
    rng_obj.mode = "Compatibility"
    initial_obj_id = id(rng_obj)  # this should never change - Singleton object
    assert neuron.h.rngMode == 0

    rng_obj.mode = "Random123"
    assert neuron.h.rngMode == 1

    rng_obj.mode = "UpdatedMCell"
    assert neuron.h.rngMode == 2

    rng_obj.mode = "Random123"
    assert neuron.h.rngMode == 1
    assert rng_obj.mode == "Random123"
    assert id(rng_obj) == initial_obj_id

    try:
        rng_obj.mode = "MersenneTwister"
    except Exception as e:
        assert isinstance(e, UndefinedRNGException)

    # make sure only one object is created
    assert rng_obj is RNGSettings.get_instance()
    assert id(rng_obj) == id(RNGSettings.get_instance()) == initial_obj_id


def test_str_repr_obj():
    """Test the str and repr methods of RNGSettings."""
    rng_obj = RNGSettings.get_instance()
    rng_obj.set_seeds()
    rng_obj.mode = "UpdatedMCell"
    assert repr(rng_obj) == "RNGSettings(mode=UpdatedMCell, base_seed=0, " \
                            "synapse_seed=0, " \
                            "ionchannel_seed=0, stimulus_seed=0, " \
                            "minis_seed=0)"

    assert str(rng_obj) == repr(rng_obj)



"""Type aliases used within the package."""

from typing_extensions import TypeAlias
from neuron import h as hoc_type

HocObjectType: TypeAlias = hoc_type   # until NEURON is typed, most NEURON types are this
NeuronRNG: TypeAlias = hoc_type
NeuronVector: TypeAlias = hoc_type
NeuronSection: TypeAlias = hoc_type
TStim: TypeAlias = hoc_type



"""Unit tests for circuit/validate.py."""
import pandas as pd
import pytest

from bluecellulab.circuit.synapse_properties import SynapseProperty
from bluecellulab.circuit.validate import check_nrrp_value


def test_check_nrrp_value():
    """Unit test for check nrrp value."""
    synapses = pd.DataFrame(data={SynapseProperty.NRRP: [15.0, 16.0]})

    check_nrrp_value(synapses)

    synapses[SynapseProperty.NRRP].loc[0] = 15.1
    with pytest.raises(ValueError):
        check_nrrp_value(synapses)

    synapses[SynapseProperty.NRRP].loc[0] = -1

    with pytest.raises(ValueError):
        check_nrrp_value(synapses)



# Copyright 2012-2024 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
import hashlib
from functools import lru_cache
import logging
from pathlib import Path
from typing import Optional

from bluepysnap.bbp import Cell as SnapCell
from bluepysnap.circuit_ids import CircuitNodeId, CircuitEdgeIds
from bluepysnap.exceptions import BluepySnapError
from bluepysnap import Circuit as SnapCircuit
import pandas as pd
from bluecellulab import circuit, neuron
from bluecellulab.circuit.circuit_access.definition import EmodelProperties
from bluecellulab.circuit import CellId, SynapseProperty
from bluecellulab.circuit.config import SimulationConfig
from bluecellulab.circuit.synapse_properties import SynapseProperties
from bluecellulab.circuit.config import SimulationConfig, SonataSimulationConfig
from bluecellulab.circuit.synapse_properties import (
    properties_from_snap,
    properties_to_snap,
)

logger = logging.getLogger(__name__)


class SonataCircuitAccess:
    """Sonata implementation of CircuitAccess protocol."""

    def __init__(self, simulation_config: str | Path | SimulationConfig) -> None:
        """Initialize SonataCircuitAccess object."""
        if isinstance(simulation_config, (str, Path)) and not Path(simulation_config).exists():
            raise FileNotFoundError(f"Circuit config file {simulation_config} not found.")

        if isinstance(simulation_config, SonataSimulationConfig):
            self.config: SimulationConfig = simulation_config
        else:
            self.config = SonataSimulationConfig(simulation_config)
        circuit_config = self.config.impl.config["network"]
        self._circuit = SnapCircuit(circuit_config)

    @property
    def available_cell_properties(self) -> set:
        return self._circuit.nodes.property_names

    def get_emodel_properties(self, cell_id: CellId) -> Optional[EmodelProperties]:
        cell_properties = self._circuit.nodes[cell_id.population_name].get(cell_id.id)
        if "@dynamics:AIS_scaler" in cell_properties:
            AIS_scaler = cell_properties["@dynamics:AIS_scaler"]
        else:
            AIS_scaler = 1.0
        if "@dynamics:soma_scaler" in cell_properties:
            soma_scaler = cell_properties["@dynamics:soma_scaler"]
        else:
            soma_scaler = 1.0

        return EmodelProperties(
            cell_properties["@dynamics:threshold_current"],
            cell_properties["@dynamics:holding_current"],
            AIS_scaler,
            soma_scaler,
        )

    def get_template_format(self) -> Optional[str]:
        return 'v6'

    def get_cell_properties(
        self, cell_id: CellId, properties: list[str] | str
    ) -> pd.Series:
        if isinstance(properties, str):
            properties = [properties]
        return self._circuit.nodes[cell_id.population_name].get(
            cell_id.id, properties=properties
        )

    @staticmethod
    def _compute_pop_ids(source: str, target: str) -> tuple[int, int]:
        """Compute the population ids from the population names."""
        def make_id(node_pop: str) -> int:
            pop_hash = hashlib.md5(node_pop.encode()).digest()
            return ((pop_hash[1] & 0x0f) << 8) + pop_hash[0]  # id: 12bit hash

        source_popid = make_id(source)
        target_popid = make_id(target)
        return source_popid, target_popid

    def get_population_ids(
        self, source_population_name: str, target_population_name: str
    ) -> tuple[int, int]:
        source_popid, target_popid = self._compute_pop_ids(
            source_population_name, target_population_name)
        return source_popid, target_popid

    def extract_synapses(
        self, cell_id: CellId, projections: Optional[list[str] | str]
    ) -> pd.DataFrame:
        """Extract the synapses.

        If projections is None, all the synapses are extracted.
        """
        snap_node_id = CircuitNodeId(cell_id.population_name, cell_id.id)
        edges = self._circuit.edges
        # select edges that are in the projections, if there are projections
        if projections is None or len(projections) == 0:
            edge_population_names = [x for x in edges]
        elif isinstance(projections, str):
            edge_population_names = [x for x in edges if edges[x].source.name == projections]
        else:
            edge_population_names = [x for x in edges if edges[x].source.name in projections]

        all_synapses_dfs: list[pd.DataFrame] = []
        for edge_population_name in edge_population_names:
            edge_population = edges[edge_population_name]
            afferent_edges: CircuitEdgeIds = edge_population.afferent_edges(snap_node_id)
            if len(afferent_edges) != 0:
                # first copy the common properties to modify them
                edge_properties: list[SynapseProperty | str] = list(
                    SynapseProperties.common
                )

                # remove optional properties if they are not present
                for optional_property in [SynapseProperty.U_HILL_COEFFICIENT,
                                          SynapseProperty.CONDUCTANCE_RATIO]:
                    if optional_property.to_snap() not in edge_population.property_names:
                        edge_properties.remove(optional_property)

                # if all plasticity props are present, add them
                if all(
                    x in edge_population.property_names
                    for x in SynapseProperties.plasticity
                ):
                    edge_properties += list(SynapseProperties.plasticity)

                snap_properties = properties_to_snap(edge_properties)
                synapses: pd.DataFrame = edge_population.get(afferent_edges, snap_properties)
                column_names = list(synapses.columns)
                synapses.columns = pd.Index(properties_from_snap(column_names))

                # make multiindex
                synapses = synapses.reset_index(drop=True)
                synapses.index = pd.MultiIndex.from_tuples(
                    zip([edge_population_name] * len(synapses), synapses.index),
                    names=["edge_name", "synapse_id"],
                )

                # add source_population_name as a column
                synapses["source_population_name"] = edges[edge_population_name].source.name

                # py-neurodamus
                dt = neuron.h.dt
                synapses[SynapseProperty.AXONAL_DELAY] = (
                    synapses[SynapseProperty.AXONAL_DELAY] / dt + 1e-5
                ).astype('i4') * dt

                if SynapseProperty.NRRP in synapses:
                    circuit.validate.check_nrrp_value(synapses)

                source_popid, target_popid = self.get_population_ids(
                    edge_population.source.name, edge_population.target.name)
                synapses = synapses.assign(
                    source_popid=source_popid, target_popid=target_popid
                )

                all_synapses_dfs.append(synapses)

        if len(all_synapses_dfs) == 0:
            return pd.DataFrame()
        else:
            return pd.concat(all_synapses_dfs)  # outer join that creates NaNs

    def target_contains_cell(self, target: str, cell_id: CellId) -> bool:
        return cell_id in self.get_target_cell_ids(target)

    @lru_cache(maxsize=1000)
    def is_valid_group(self, group: str) -> bool:
        return group in self._circuit.node_sets

    @lru_cache(maxsize=16)
    def get_target_cell_ids(self, target: str) -> set[CellId]:
        ids = self._circuit.nodes.ids(target)
        return {CellId(x.population, x.id) for x in ids}

    @lru_cache(maxsize=100)
    def fetch_cell_info(self, cell_id: CellId) -> pd.Series:
        return self._circuit.nodes[cell_id.population_name].get(cell_id.id)

    def fetch_mini_frequencies(self, cell_id: CellId) -> tuple[float | None, float | None]:
        cell_info = self.fetch_cell_info(cell_id)
        exc_mini_frequency = cell_info['exc-mini_frequency'] \
            if 'exc-mini_frequency' in cell_info else None
        inh_mini_frequency = cell_info['inh-mini_frequency'] \
            if 'inh-mini_frequency' in cell_info else None
        return exc_mini_frequency, inh_mini_frequency

    @property
    def node_properties_available(self) -> bool:
        return True

    def get_gids_of_mtypes(self, mtypes: list[str]) -> set[CellId]:
        all_cell_ids = set()
        all_population_names: list[str] = list(self._circuit.nodes)
        for population_name in all_population_names:
            try:
                cell_ids = self._circuit.nodes[population_name].ids(
                    {SnapCell.MTYPE: mtypes})
            except BluepySnapError:
                continue
            all_cell_ids |= {CellId(population_name, id) for id in cell_ids}
        return all_cell_ids

    def get_cell_ids_of_targets(self, targets: list[str]) -> set[CellId]:
        cell_ids = set()
        for target in targets:
            cell_ids |= self.get_target_cell_ids(target)
        return cell_ids

    def morph_filepath(self, cell_id: CellId) -> str:
        """Returns the .asc morphology path from 'alternate_morphologies'."""
        node_population = self._circuit.nodes[cell_id.population_name]
        try:  # if asc defined in alternate morphology
            return str(node_population.morph.get_filepath(cell_id.id, extension="asc"))
        except BluepySnapError as e:
            logger.debug(f"No asc morphology found for {cell_id}, trying swc.")
            return str(node_population.morph.get_filepath(cell_id.id))

    def emodel_path(self, cell_id: CellId) -> str:
        node_population = self._circuit.nodes[cell_id.population_name]
        return str(node_population.models.get_filepath(cell_id.id))



# Copyright 2012-2024 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import setuptools


# Read the README.rst file
with open("README.rst", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="bluecellulab",
    use_scm_version={
        'version_scheme': 'python-simplified-semver',
        'local_scheme': 'no-local-version'
    },
    python_requires='>=3.8',
    setup_requires=['setuptools_scm'],
    packages=setuptools.find_packages(include=['bluecellulab', 'bluecellulab.*']),
    author="Blue Brain Project, EPFL",
    description="The Pythonic Blue Brain simulator access",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    license="Apache2.0",
    install_requires=[
        "NEURON>=8.0.2,<9.0.0",
        "numpy>=1.8.0,<2.0.0",
        "matplotlib>=3.0.0,<4.0.0",
        "pandas>=1.0.0,<3.0.0",
        "bluepysnap>=3.0.0,<4.0.0",
        "pydantic>=2.5.2,<3.0.0",
        "typing-extensions>=4.8.0",
        "networkx>=3.1"
    ],
    keywords=[
        'computational neuroscience',
        'simulation',
        'analysis',
        'parameters',
        'Blue Brain Project'],
    url="https://github.com/BlueBrain/BlueCelluLab",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX',
        'Topic :: Scientific/Engineering',
        'Programming Language :: Python :: 3',
        'Topic :: Utilities',
    ],
    include_package_data=True,
)



"""Unit tests for BGLib functionality"""

import os

import bluecellulab

script_dir = os.path.dirname(__file__)


class TestBGLibBaseClass1:

    """Test BGLib base class"""

    def setup_method(self):
        """Setup"""
        self.taper_cell = bluecellulab.Cell(
            "%s/examples/tapertest_cells/taper_cell.hoc" % script_dir,
            "%s/examples/tapertest_cells" % script_dir)
        self.notaper_cell = bluecellulab.Cell(
            "%s/examples/tapertest_cells/notaper_cell.hoc" % script_dir,
            "%s/examples/tapertest_cells" % script_dir)

    def teardown_method(self):
        """Teardown"""
        del self.taper_cell
        del self.notaper_cell

    def test_tapering(self):
        """BGLib: Test if tapering in Cell.hoc works correctly"""
        import numpy as np
        axon0_taper_diams = [segment.diam
                             for segment in self.taper_cell.axonal[0]]
        np.testing.assert_array_almost_equal(
            axon0_taper_diams, [
                1.9, 1.7, 1.5, 1.3, 1.1])

        axon1_taper_diams = [segment.diam
                             for segment in self.taper_cell.axonal[1]]
        np.testing.assert_array_almost_equal(
            axon1_taper_diams,
            [0.9642857142857143, 0.8928571428571429, 0.8214285714285714, 0.75,
             0.6785714285714286, 0.6071428571428571, 0.5357142857142857])

        myelin_taper_diams = [
            segment.diam
            for segment in self.taper_cell.cell.getCell().myelin[0]]
        np.testing.assert_array_almost_equal(
            myelin_taper_diams,
            [0.25] *
            21)

        axon0_notaper_diams = [segment.diam
                               for segment in self.notaper_cell.axonal[0]]
        np.testing.assert_array_almost_equal(
            axon0_notaper_diams,
            [0.7484678706370597])

        axon1_notaper_diams = [segment.diam
                               for segment in self.notaper_cell.axonal[1]]
        np.testing.assert_array_almost_equal(
            axon1_notaper_diams,
            [0.14187096846614167])



import logging
import os
from pathlib import Path
import pytest
from types import ModuleType
from unittest.mock import MagicMock, patch
from bluecellulab import importer
from bluecellulab.exceptions import BluecellulabError


@patch("os.path.isdir", return_value=False)  # when x86_64 isdir returns False
def test_import_mod_lib_no_env_no_folder(mocked_isdir):
    mock_neuron = MagicMock()
    with patch.dict(os.environ, {}, clear=True):
        assert importer.import_mod_lib(mock_neuron) == "No mechanisms are loaded."


def test_import_mod_lib_env_var_set_folder_exists():
    mock_neuron = MagicMock()
    with patch.dict(os.environ, {"BLUECELLULAB_MOD_LIBRARY_PATH": "/fake/path"}):
        with patch(
            "os.path.isdir", return_value=True
        ):  # when x86_64 isdir returns True and env var is set
            with pytest.raises(
                BluecellulabError,
                match="BLUECELLULAB_MOD_LIBRARY_PATH is set and current directory contains the x86_64 folder. Please remove one of them.",
            ):
                importer.import_mod_lib(mock_neuron)


def test_import_mod_lib_env_var_set():
    mock_neuron = MagicMock()
    with patch.dict(os.environ, {"BLUECELLULAB_MOD_LIBRARY_PATH": "/fake/path"}):
        with patch("os.path.isdir", return_value=False):
            assert importer.import_mod_lib(mock_neuron) == "/fake/path"


def test_import_mod_lib_so_file():
    mock_neuron = MagicMock()
    fake_so_path = "/fake/path/to/library.so"
    with patch.dict(os.environ, {"BLUECELLULAB_MOD_LIBRARY_PATH": fake_so_path}):
        with patch("os.path.isdir", return_value=False):
            importer.import_mod_lib(mock_neuron)
            mock_neuron.h.nrn_load_dll.assert_called_with(fake_so_path)


def test_import_mod_lib_no_env_with_folder():
    mock_neuron = MagicMock()
    with patch.dict(os.environ, {}, clear=True):
        with patch("os.path.isdir", return_value=True):
            assert importer.import_mod_lib(mock_neuron).endswith("x86_64")


@patch.object(
    importer.resources,
    "files",
    return_value=Path("/fake/path/"),
)
def test_import_neurodamus(mocked_resources):
    mock_neuron = MagicMock()
    importer.import_neurodamus(mock_neuron)
    assert mock_neuron.h.load_file.called
    # Check that it was called with the expected arguments
    mock_neuron.h.load_file.assert_any_call("/fake/path/hoc/Cell.hoc")


def test_print_header(caplog):
    # Creating a dummy ModuleType object with an attribute '__file__'
    dummy_neuron = ModuleType("dummy_neuron")
    dummy_neuron.__file__ = "/path/to/neuron"

    mod_lib_path = "/path/to/mod_lib"
    with caplog.at_level(logging.DEBUG):
        importer.print_header(dummy_neuron, mod_lib_path)

    assert "Imported NEURON from: /path/to/neuron" in caplog.text
    assert "Mod lib: /path/to/mod_lib" in caplog.text


def test_print_header_with_decorator(caplog):
    """Ensure the decorator loading hoc and mod files work as expected."""
    with caplog.at_level(logging.DEBUG):
        @importer.load_hoc_and_mod_files
        def x():
            pass

        x()  # call 3 times to ensure the decorator is called only once
        x()
        x()

    assert caplog.text.count("Loading the mod files.") == 1
    assert caplog.text.count("Loading the hoc files.") == 1



"""Testing the loading of the bluecellulab"""

import pytest
import logging
import io


@pytest.mark.unit
def test_moduleload():
    """bluecellulab: Test the loading of the bluecellulab module"""
    import bluecellulab  # NOQA


@pytest.mark.unit
def test_logger():
    """Test logging"""

    import bluecellulab  # NOQA

    # Verify the hierarchy and level inheritance
    parent_logger = logging.getLogger('bluecellulab')
    parent_logger.setLevel(logging.DEBUG)
    child_logger = logging.getLogger('bluecellulab.' + __name__)
    assert child_logger.getEffectiveLevel() == parent_logger.getEffectiveLevel()
    assert child_logger.parent is parent_logger

    # Verify that a warning logger entry is not displayed when the logger level is set to critical
    logger = logging.getLogger('test_logger')
    logger.setLevel(logging.CRITICAL)
    stream = io.StringIO()
    handler = logging.StreamHandler(stream)
    logger.addHandler(handler)
    logger.warning('This is a warning')
    log_output = stream.getvalue()
    assert log_output == ''

    # Verify that a warning logger entry is displayed when the logger level is set to warning
    logger.setLevel(logging.WARNING)
    logger.warning('This is a warning')
    log_output = stream.getvalue()
    assert log_output == 'This is a warning\n'

    logger.removeHandler(handler)



"""Unit tests for template.py module."""

from pathlib import Path
from unittest.mock import Mock

import pytest

from bluecellulab.cell.template import NeuronTemplate, public_hoc_cell
from bluecellulab.circuit.circuit_access import EmodelProperties
from bluecellulab.exceptions import BluecellulabError


examples_dir = Path(__file__).resolve().parent.parent / "examples"

hipp_hoc_path = (
    examples_dir
    / "hippocampus_opt_cell_template"
    / "electrophysiology"
    / "cell.hoc"
)
hipp_morph_path = (
    examples_dir / "hippocampus_opt_cell_template" / "morphology" / "cell.asc"
)

v6_hoc_path = (
    examples_dir / "circuit_sonata_quick_scx" / "components" / "hoc" / "cADpyr_L2TPC.hoc"
)

v6_morph_path = (
    examples_dir / "circuit_sonata_quick_scx" / "components" / "morphologies" / "asc" / "rr110330_C3_idA.asc"
)


def test_get_cell_with_bluepyopt_template():
    """Unit test for the get_cell method with bluepyopt_template."""
    template = NeuronTemplate(hipp_hoc_path, hipp_morph_path, "bluepyopt", None)
    cell = template.get_cell(gid=None)
    assert cell.hname() == f"bACnoljp_bluecellulab_{(hex(id(template)))}[0]"


def test_neuron_template_init():
    """Unit test for the NeuronTemplate's constructor."""
    missing_file = "missing_file"

    with pytest.raises(FileNotFoundError):
        NeuronTemplate(missing_file, hipp_morph_path, "bluepyopt", None)
    with pytest.raises(FileNotFoundError):
        NeuronTemplate(hipp_hoc_path, missing_file, "bluepyopt", None)

    NeuronTemplate(hipp_hoc_path, hipp_morph_path, "bluepyopt", None)


def test_public_hoc_cell_bluepyopt_template():
    """Unit test for public_hoc_cell."""
    template = NeuronTemplate(hipp_hoc_path, hipp_morph_path, "bluepyopt", None)
    cell = template.get_cell(None)
    hoc_public = public_hoc_cell(cell)
    assert hoc_public.gid == 0.0


def test_public_hoc_cell_v6_template():
    """Unit test for public_hoc_cell."""
    emodel_properties = EmodelProperties(
        threshold_current=1.1433533430099487,
        holding_current=1.4146618843078613,
        AIS_scaler=1.4561502933502197,
        soma_scaler=1.0
    )
    template = NeuronTemplate(v6_hoc_path, v6_morph_path, "v6", emodel_properties)
    cell = template.get_cell(5)
    hoc_public = public_hoc_cell(cell)
    assert hoc_public.gid == 5.0


def test_public_hoc_cell_failure():
    """Unit test for public_hoc_cell when neither getCell nor CellRef is provided."""
    cell_without_getCell_or_CellRef = Mock(spec=[])  # spec=[] ensures no attributes exist
    with pytest.raises(BluecellulabError) as excinfo:
        public_hoc_cell(cell_without_getCell_or_CellRef)
    assert "Public cell properties cannot be accessed" in str(excinfo.value)


def test_load_bpo_template():
    """Test the loading of a hoc without getCell or gid."""
    hoc_path = examples_dir / "bpo_cell" / "0_cADpyr_L5TPC_a6e707a_1_sNone.hoc"
    morph_path = examples_dir / "bpo_cell" / "C060114A5.asc"
    neuron_template = NeuronTemplate(hoc_path, morph_path, "bluepyopt", None)
    cell = neuron_template.get_cell(None)
    assert len(cell.soma[0].children()) == 11



"""Unit tests for the stimuli module."""

from pydantic import ValidationError
import pytest
from bluecellulab.stimulus.circuit_stimulus_definitions import SynapseReplay


def test_synapse_replay_validator():
    """Assures synapse replay's validator fails."""
    with pytest.raises(ValidationError):
        _ = SynapseReplay(
            target="target1",
            source="source1",
            delay=0,
            duration=3000,
            spike_file="file_that_does_not_exist.dat",
        )



# Copyright 2012-2024 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Class that represents a synapse in bluecellulab."""

from __future__ import annotations
from typing import Any, NamedTuple, Optional
import logging

import neuron
import pandas as pd

from bluecellulab.circuit import SynapseProperty
from bluecellulab.circuit.node_id import CellId
from bluecellulab.rngsettings import RNGSettings
from bluecellulab.type_aliases import HocObjectType, NeuronSection


logger = logging.getLogger(__name__)


class SynapseID(NamedTuple):
    """Class that represents a synapse id in bluecellulab."""

    projection: str
    sid: int


class SynapseHocArgs(NamedTuple):
    """Parameters required by synapse hoc constructor."""
    location: float
    section: NeuronSection


class Synapse:
    """Class that represents a synapse in bluecellulab."""

    def __init__(
            self,
            cell_id: CellId,
            hoc_args: SynapseHocArgs,
            syn_id: tuple[str, int],
            syn_description: pd.Series,
            popids: tuple[int, int],
            extracellular_calcium: float | None = None):
        """Constructor.

        Args:
            gid: The post-synaptic cell gid.
            hoc_args: The synapse location and section in hoc.
            syn_id: A tuple containing a synapse identifier, where the string is
                the projection name and int is the synapse id. An empty string
                refers to a local connection.
            syn_description: Parameters of the synapse.
            popids: A tuple containing source and target popids used by the random
                number generation.
            extracellular_calcium: The extracellular calcium concentration. Optional
                and defaults to None.
        """
        self.persistent: list[HocObjectType] = []
        self.synapseconfigure_cmds: list[str] = []
        self._delay_weights: list[tuple[float, float]] = []
        self._weight: Optional[float] = None

        self.post_cell_id = cell_id
        self.syn_id = SynapseID(*syn_id)
        self.extracellular_calcium = extracellular_calcium
        self.syn_description: pd.Series = self.update_syn_description(syn_description)
        self.hsynapse: Optional[HocObjectType] = None

        self.source_popid, self.target_popid = popids

        self.pre_gid = int(self.syn_description[SynapseProperty.PRE_GID])

        self.hoc_args = hoc_args
        self.mech_name: str = "not-yet-defined"
        self.randseed3: Optional[int] = None

    @property
    def delay_weights(self) -> list[tuple[float, float]]:
        """Adjustments to synapse delay and weight."""
        return self._delay_weights

    @delay_weights.setter
    def delay_weights(self, value: list[tuple[float, float]]) -> None:
        self._delay_weights = value

    @property
    def weight(self) -> float | None:
        """The last overridden synapse weight."""
        return self._weight

    @weight.setter
    def weight(self, value: float | None) -> None:
        self._weight = value

    def update_syn_description(self, syn_description: pd.Series) -> pd.Series:
        """Change data types, compute more columns needed by the simulator."""
        # if the optional properties are NaN (that happens due to pandas outer join), then remove them
        for prop in [SynapseProperty.U_HILL_COEFFICIENT, SynapseProperty.NRRP]:
            if prop in syn_description and pd.isna(syn_description[prop]):
                syn_description.pop(prop)

        if SynapseProperty.NRRP in syn_description:
            try:
                int(syn_description[SynapseProperty.NRRP])
            except ValueError:
                # delete SynapseProperty.NRRP from syn_description
                syn_description.pop(SynapseProperty.NRRP)

        if SynapseProperty.U_HILL_COEFFICIENT in syn_description:
            syn_description["u_scale_factor"] = self.calc_u_scale_factor(
                syn_description[SynapseProperty.U_HILL_COEFFICIENT], self.extracellular_calcium)
        else:
            syn_description["u_scale_factor"] = 1.0
        syn_description[SynapseProperty.U_SYN] *= syn_description["u_scale_factor"]
        return syn_description

    def apply_hoc_configuration(self, hoc_configure_params: list[str]) -> None:
        """Apply the list of hoc configuration commands to the synapse."""
        self.synapseconfigure_cmds = []
        # hoc exec synapse configure blocks
        for cmd in hoc_configure_params:
            self.synapseconfigure_cmds.append(cmd)
            cmd = cmd.replace('%s', '\n%(syn)s')
            hoc_cmd = cmd % {'syn': self.hsynapse.hname()}  # type: ignore
            hoc_cmd = '{%s}' % hoc_cmd
            neuron.h(hoc_cmd)

    def _set_gabaab_ampanmda_rng(self) -> None:
        """Setup the RNG for the gabaab and ampanmd helpers.

        Raises:
            ValueError: when rng mode is not recognised.
        """
        rng_settings = RNGSettings.get_instance()
        if rng_settings.mode == "Random123":
            self.randseed1 = self.post_cell_id.id + 250
            self.randseed2 = self.syn_id.sid + 100
            self.randseed3 = self.source_popid * 65536 + self.target_popid + \
                rng_settings.synapse_seed + 300
            self.hsynapse.setRNG(  # type: ignore
                self.randseed1,
                self.randseed2,
                self.randseed3)
        else:
            rndd = neuron.h.Random()
            if rng_settings.mode == "Compatibility":
                self.randseed1 = self.syn_id.sid * 100000 + 100
                self.randseed2 = self.post_cell_id.id + \
                    250 + rng_settings.base_seed
            elif rng_settings.mode == "UpdatedMCell":
                self.randseed1 = self.syn_id.sid * 1000 + 100
                self.randseed2 = self.source_popid * 16777216 + \
                    self.post_cell_id.id + \
                    250 + rng_settings.base_seed + \
                    rng_settings.synapse_seed
            else:
                raise ValueError(
                    "Synapse: unknown RNG mode: %s" %
                    rng_settings.mode)
            self.randseed3 = None  # Not used in this case
            rndd.MCellRan4(self.randseed1, self.randseed2)
            rndd.uniform(0, 1)
            self.hsynapse.setRNG(rndd)  # type: ignore
            self.persistent.append(rndd)

    def delete(self) -> None:
        """Delete the NEURON objects of the connection."""
        if hasattr(self, 'persistent'):
            for persistent_object in self.persistent:
                del persistent_object

    @staticmethod
    def calc_u_scale_factor(u_hill_coefficient, extracellular_calcium):
        if extracellular_calcium is None or u_hill_coefficient is None:
            return 1.0

        def constrained_hill(K_half, y) -> float:
            """Calculates the constrained hill coefficient."""
            K_half_fourth = K_half**4
            y_fourth = y**4
            return (K_half_fourth + 16) / 16 * y_fourth / (K_half_fourth + y_fourth)

        u_scale_factor = constrained_hill(u_hill_coefficient, extracellular_calcium)
        logger.debug(
            "Scaling synapse Use with u_hill_coeffient %f, "
            "extra_cellular_calcium %f with a factor of %f" %
            (u_hill_coefficient, extracellular_calcium, u_scale_factor))

        return u_scale_factor

    @property
    def info_dict(self) -> dict[str, Any]:
        """Convert the synapse info to a dict from which it can be
        reconstructed."""
        synapse_dict: dict[str, Any] = {}

        synapse_dict['synapse_id'] = self.syn_id
        synapse_dict['pre_cell_id'] = self.pre_gid
        synapse_dict['post_cell_id'] = self.post_cell_id.id
        synapse_dict['syn_description'] = self.syn_description.to_dict()
        # if keys are enum make them str
        synapse_dict['syn_description'] = {
            str(k): v for k, v in synapse_dict['syn_description'].items()}

        synapse_dict['post_segx'] = self.hoc_args.location
        synapse_dict['mech_name'] = self.mech_name
        synapse_dict['randseed1'] = self.randseed1
        synapse_dict['randseed2'] = self.randseed2
        synapse_dict['randseed3'] = self.randseed3
        synapse_dict['synapseconfigure_cmds'] = self.synapseconfigure_cmds

        # Parameters of the mod mechanism
        synapse_dict['synapse_parameters'] = {}
        synapse_dict['synapse_parameters']['Use'] = self.hsynapse.Use  # type: ignore
        synapse_dict['synapse_parameters']['Dep'] = self.hsynapse.Dep  # type: ignore
        synapse_dict['synapse_parameters']['Fac'] = self.hsynapse.Fac  # type: ignore

        synapse_dict['synapse_parameters']['extracellular_calcium'] = \
            self.extracellular_calcium

        return synapse_dict

    def __del__(self) -> None:
        self.delete()


class GluSynapse(Synapse):

    def __init__(self, gid, hoc_args, syn_id, syn_description, popids, extracellular_calcium):
        super().__init__(gid, hoc_args, syn_id, syn_description, popids, extracellular_calcium)
        self.use_glusynapse_helper()

    def use_glusynapse_helper(self) -> None:
        """Python implementation of the GluSynapse helper.

        This helper object will encapsulate the hoc actions  needed to
        create our plastic excitatory synapse.
        """
        self.mech_name = 'GluSynapse'

        self.hsynapse = neuron.h.GluSynapse(
            self.hoc_args.location,
            sec=self.hoc_args.section
        )

        self.hsynapse.Use_d = self.syn_description["Use_d_TM"] * \
            self.syn_description["u_scale_factor"]
        self.hsynapse.Use_p = self.syn_description["Use_p_TM"] * \
            self.syn_description["u_scale_factor"]

        self.hsynapse.theta_d_GB = self.syn_description["theta_d"]
        self.hsynapse.theta_p_GB = self.syn_description["theta_p"]
        self.hsynapse.rho0_GB = self.syn_description["rho0_GB"]

        self.hsynapse.volume_CR = self.syn_description["volume_CR"]

        self.hsynapse.gmax_d_AMPA = self.syn_description["gmax_d_AMPA"]
        self.hsynapse.gmax_p_AMPA = self.syn_description["gmax_p_AMPA"]

        if self.hsynapse.rho0_GB > neuron.h.rho_star_GB_GluSynapse:
            self.hsynapse.gmax0_AMPA = self.hsynapse.gmax_p_AMPA
            self.hsynapse.Use = self.hsynapse.Use_p
        else:
            self.hsynapse.gmax0_AMPA = self.hsynapse.gmax_d_AMPA
            self.hsynapse.Use = self.hsynapse.Use_d

        self.hsynapse.gmax_NMDA = self.hsynapse.gmax0_AMPA * \
            self.syn_description[SynapseProperty.CONDUCTANCE_RATIO]
        self.hsynapse.tau_d_AMPA = self.syn_description[SynapseProperty.DTC]
        self.hsynapse.Dep = abs(self.syn_description[SynapseProperty.D_SYN])
        self.hsynapse.Fac = abs(self.syn_description[SynapseProperty.F_SYN])

        if self.syn_description[SynapseProperty.NRRP] >= 0:
            self.hsynapse.Nrrp = self.syn_description[SynapseProperty.NRRP]

        self.randseed1 = self.post_cell_id.id
        self.randseed2 = 100000 + self.syn_id.sid
        rng_settings = RNGSettings.get_instance()
        self.randseed3 = rng_settings.synapse_seed + 200
        self.hsynapse.setRNG(self.randseed1, self.randseed2, self.randseed3)
        self.hsynapse.synapseID = self.syn_id.sid

    @property
    def info_dict(self):
        parent_dict = super().info_dict
        parent_dict['synapse_parameters']['tau_d_AMPA'] = self.hsynapse.tau_d_AMPA
        return parent_dict


class GabaabSynapse(Synapse):

    def __init__(self, gid, hoc_args, syn_id, syn_description, popids, extracellular_calcium, randomize_risetime=True):
        super().__init__(gid, hoc_args, syn_id, syn_description, popids, extracellular_calcium)
        self.use_gabaab_helper(randomize_risetime)

    def use_gabaab_helper(self, randomize_gaba_risetime: bool) -> None:
        """Python implementation of the GABAAB helper.

        This helper object will encapsulate the hoc actions
         needed to create our typical inhibitory synapse.

        Args:
            randomize_gaba_risetime: if True then RNG code runs
            to set tau_r_GABAA value.

        Raises:
            ValueError: raised when the RNG mode is unknown.
        """
        self.mech_name = 'ProbGABAAB_EMS'

        self.hsynapse = neuron.h.ProbGABAAB_EMS(
            self.hoc_args.location,
            sec=self.hoc_args.section
        )

        if randomize_gaba_risetime is True:
            rng = neuron.h.Random()
            rng_settings = RNGSettings.get_instance()
            if rng_settings.mode == "Compatibility":
                rng.MCellRan4(
                    self.syn_id.sid * 100000 + 100,
                    self.post_cell_id.id + 250 + rng_settings.base_seed)
            elif rng_settings.mode == "UpdatedMCell":
                rng.MCellRan4(
                    self.syn_id.sid * 1000 + 100,
                    self.source_popid *
                    16777216 +
                    self.post_cell_id.id +
                    250 +
                    rng_settings.base_seed +
                    rng_settings.synapse_seed)
            elif rng_settings.mode == "Random123":
                rng.Random123(
                    self.post_cell_id.id +
                    250,
                    self.syn_id.sid +
                    100,
                    self.source_popid *
                    65536 +
                    self.target_popid +
                    rng_settings.synapse_seed +
                    450)
            else:
                raise ValueError(
                    "Synapse: unknown RNG mode: %s" %
                    rng_settings.mode)

            rng.lognormal(0.2, 0.1)
            self.hsynapse.tau_r_GABAA = rng.repick()

        if SynapseProperty.CONDUCTANCE_RATIO in self.syn_description:
            self.hsynapse.GABAB_ratio = self.syn_description[SynapseProperty.CONDUCTANCE_RATIO]

        self.hsynapse.tau_d_GABAA = self.syn_description[SynapseProperty.DTC]
        self.hsynapse.Use = abs(self.syn_description[SynapseProperty.U_SYN])
        self.hsynapse.Dep = abs(self.syn_description[SynapseProperty.D_SYN])
        self.hsynapse.Fac = abs(self.syn_description[SynapseProperty.F_SYN])

        if SynapseProperty.NRRP in self.syn_description:
            self.hsynapse.Nrrp = self.syn_description[SynapseProperty.NRRP]

        self._set_gabaab_ampanmda_rng()

        self.hsynapse.synapseID = self.syn_id.sid

    @property
    def info_dict(self):
        parent_dict = super().info_dict
        parent_dict['synapse_parameters']['tau_d_GABAA'] = self.hsynapse.tau_d_GABAA
        parent_dict['synapse_parameters']['tau_r_GABAA'] = self.hsynapse.tau_r_GABAA
        return parent_dict


class AmpanmdaSynapse(Synapse):

    def __init__(self, gid, hoc_args, syn_id, syn_description, popids, extracellular_calcium):
        super().__init__(gid, hoc_args, syn_id, syn_description, popids, extracellular_calcium)
        self.use_ampanmda_helper()

    def use_ampanmda_helper(self) -> None:
        """Python implementation of the AMPANMDA helper.

        This helper object will encapsulate the hoc actions  needed to
        create our typical excitatory synapse.
        """
        self.mech_name = 'ProbAMPANMDA_EMS'

        self.hsynapse = neuron.h.ProbAMPANMDA_EMS(
            self.hoc_args.location,
            sec=self.hoc_args.section,
        )
        self.hsynapse.tau_d_AMPA = self.syn_description[SynapseProperty.DTC]
        if SynapseProperty.CONDUCTANCE_RATIO in self.syn_description:
            self.hsynapse.NMDA_ratio = self.syn_description[SynapseProperty.CONDUCTANCE_RATIO]

        self.hsynapse.Use = abs(self.syn_description[SynapseProperty.U_SYN])
        self.hsynapse.Dep = abs(self.syn_description[SynapseProperty.D_SYN])
        self.hsynapse.Fac = abs(self.syn_description[SynapseProperty.F_SYN])

        if SynapseProperty.NRRP in self.syn_description:
            self.hsynapse.Nrrp = self.syn_description[SynapseProperty.NRRP]

        self._set_gabaab_ampanmda_rng()
        self.hsynapse.synapseID = self.syn_id.sid

    @property
    def info_dict(self):
        parent_dict = super().info_dict
        parent_dict['synapse_parameters']['tau_d_AMPA'] = self.hsynapse.tau_d_AMPA
        return parent_dict



"""Unit tests for the synapse module."""

import numpy as np
from pytest import approx

from bluecellulab.synapse import Synapse


def test_calc_u_scale_factor():
    """Test calculation of u_scale_factor against its old implementation."""
    def hill(extracellular_calcium, y, K_half):
        return y * extracellular_calcium**4 / (
            K_half**4 + extracellular_calcium**4)

    def constrained_hill(K_half):
        y_max = (K_half**4 + 16) / 16
        return lambda x: hill(x, y_max, K_half)

    def f_scale(x, y):
        return constrained_hill(x)(y)

    scale_factor = np.vectorize(f_scale)

    a = 4.62799366
    b = 3.27495564

    assert scale_factor(a, 2) == approx(Synapse.calc_u_scale_factor(a, 2))
    assert scale_factor(a, 2.2) == approx(Synapse.calc_u_scale_factor(a, 2.2))
    assert scale_factor(a, b) == approx(Synapse.calc_u_scale_factor(a, b))



# Copyright 2012-2024 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module that handles the global NEURON parameters."""

import neuron
from bluecellulab.circuit.config.sections import Conditions, MechanismConditions
from bluecellulab.exceptions import error_context


def set_global_condition_parameters(condition_parameters: Conditions) -> None:
    """Sets the global condition parameters in NEURON objects if GluSynapse is
    available."""
    if condition_parameters.extracellular_calcium is not None and hasattr(
        neuron.h, "cao_CR_GluSynapse"
    ):
        cao_cr_glusynapse = condition_parameters.extracellular_calcium
        with error_context("mechanism/s for cao_CR_GluSynapse need to be compiled"):
            neuron.h.cao_CR_GluSynapse = cao_cr_glusynapse

    mechanism_conditions = condition_parameters.mech_conditions
    if mechanism_conditions is not None:
        set_minis_single_vesicle_values(mechanism_conditions)
        set_init_depleted_values(mechanism_conditions)


def set_init_depleted_values(mech_conditions: MechanismConditions) -> None:
    """Set the init_depleted values in NEURON."""
    with error_context("mechanism/s for init_depleted need to be compiled"):
        if mech_conditions.glusynapse and mech_conditions.glusynapse.init_depleted is not None:
            neuron.h.init_depleted_GluSynapse = mech_conditions.glusynapse.init_depleted
        if mech_conditions.ampanmda and mech_conditions.ampanmda.init_depleted is not None:
            neuron.h.init_depleted_ProbAMPANMDA_EMS = mech_conditions.ampanmda.init_depleted
        if mech_conditions.gabaab and mech_conditions.gabaab.init_depleted is not None:
            neuron.h.init_depleted_ProbGABAAB_EMS = mech_conditions.gabaab.init_depleted


def set_minis_single_vesicle_values(mech_conditions: MechanismConditions) -> None:
    """Set the minis_single_vesicle values in NEURON."""
    with error_context("mechanism/s for minis_single_vesicle need to be compiled"):
        if mech_conditions.ampanmda and mech_conditions.ampanmda.minis_single_vesicle is not None:
            neuron.h.minis_single_vesicle_ProbAMPANMDA_EMS = (
                mech_conditions.ampanmda.minis_single_vesicle
            )
        if mech_conditions.gabaab and mech_conditions.gabaab.minis_single_vesicle is not None:
            neuron.h.minis_single_vesicle_ProbGABAAB_EMS = (
                mech_conditions.gabaab.minis_single_vesicle
            )
        if mech_conditions.glusynapse and mech_conditions.glusynapse.minis_single_vesicle is not None:
            neuron.h.minis_single_vesicle_GluSynapse = (
                mech_conditions.glusynapse.minis_single_vesicle
            )


def set_tstop_value(tstop: float) -> None:
    """Set the tstop value required by Tstim noise stimuli."""
    neuron.h.tstop = tstop
